try:
    from deep_utils.dummy_objects.vision.face_detection import HaarcascadeCV2FaceDetector
    from .cv2_.haarcascade_cv2_face_detection import HaarcascadeCV2FaceDetector
except ModuleNotFoundError:
    pass



def download_binary(url: str):
    import urllib.request
    from io import BytesIO
    if not url.startswith('http'):
        url = "http://" + url
    response = urllib.request.urlopen(url)
    data = response.read()

    file = BytesIO(data)
    return file



class Config:
    def __init__(self):
        self.device = "cpu"
        self.min_face_size = 20.0
        self.thresholds = (0.6, 0.7, 0.8)
        self.min_detection_size = 12
        self.factor = 0.707
        self.top_k = 5000
        self.nms_thresholds = 0.4
        self.keep_top_k = 750
        self.visualization_threshold = 0.6
        self.resnet50_url = "https://github.com/Practical-AI/deep_utils/releases/download/0.2.0/resnet50.pth"
        self.resnet50_cache = (
            "weights/vision/face_detection/retinaface/torch/resnet50.pth"
        )
        self.resnet50 = None
        self.mobilenet_url = "https://github.com/Practical-AI/deep_utils/releases/download/0.2.0/mobilenet.pth"
        self.mobilenet_cache = (
            "weights/vision/face_detection/retinaface/torch/mobilenet.pth"
        )
        self.mobilenet = None
        self.mobilenetV1X_url = "https://github.com/Practical-AI/deep_utils/releases/download/0.2.0/mobilenetV1X0.25_pretrain.tar"
        self.mobilenetV1X_cache = "weights/vision/face_detection/retinaface/torch/mobilenetV1X0.25_pretrain.tar"
        self.mobilenetV1X = None
        self.network = "mobilenet"
        self.cfg_mnet = {
            "name": "mobilenet",
            "min_sizes": [[16, 32], [64, 128], [256, 512]],
            "steps": [8, 16, 32],
            "variance": [0.1, 0.2],
            "clip": False,
            "loc_weight": 2.0,
            "gpu_train": True,
            "batch_size": 32,
            "ngpu": 1,
            "epoch": 250,
            "decay1": 190,
            "decay2": 220,
            "image_size": 640,
            "pretrain": True,
            "return_layers": {"stage1": 1, "stage2": 2, "stage3": 3},
            "in_channel": 32,
            "out_channel": 64,
        }

        self.cfg_re50 = {
            "name": "resnet50",
            "min_sizes": [[16, 32], [64, 128], [256, 512]],
            "steps": [8, 16, 32],
            "variance": [0.1, 0.2],
            "clip": False,
            "loc_weight": 2.0,
            "gpu_train": True,
            "batch_size": 24,
            "ngpu": 4,
            "epoch": 100,
            "decay1": 70,
            "decay2": 90,
            "image_size": 840,
            "pretrain": True,
            "return_layers": {"layer2": 1, "layer3": 2, "layer4": 3},
            "in_channel": 256,
            "out_channel": 256,
        }



import os.path
from pathlib import Path
from typing import Union, List
import numpy as np
import torch
from PIL import Image
import cv2
from .crnn_model import CRNNModelTorch
from deep_utils.utils.ctc_decoder.ctc_decoder import CTCDecoder
from deep_utils.vision.vision_utils.torch_vision_utils.torch_vision_utils import TorchVisionUtils


class CRNNInferenceTorch:
    def __init__(self, model_path, decode_method='greedy', device='cpu'):
        self.device = device
        self.decode_method = decode_method
        state_dict = torch.load(model_path, map_location=self.device)
        self.model = CRNNModelTorch(state_dict['img_h'], n_channels=state_dict['n_channels'],
                                    n_classes=state_dict['n_classes'], n_hidden=state_dict['n_hidden'],
                                    lstm_input=state_dict['lstm_input'])
        try:
            self.model.load_state_dict(state_dict["state_dict"])
        except:
            self.model.load_state_dict(
                {
                    ".".join(k.split(".")[1:]): v
                    for k, v in state_dict["state_dict"].items()
                }
            )
        self.model.eval()
        self.model.to(self.device)

        self.label2char = state_dict['label2char']
        self.transformer = state_dict['val_transform']
        del state_dict

    def infer(self, img: Union[str, Path, np.ndarray], get_string=True, mean_min_conf: int = 0):
        """
        :param img:
        :param get_string: Get the string
        :param mean_min_conf: If set to a value larger than zero, returns the mean of the least probabilities.
        It's a good indicator for showing mode's strength and the confidence of the model!
        :return:
        """
        if isinstance(img, np.ndarray):
            pass
        elif os.path.isfile(img):
            img = cv2.imread(img)[..., ::-1]
        else:
            raise ValueError(f"The input image:{img} is invalid")
        image = self.transformer(image=img)['image'][0:1, ...]
        image = image.view(1, *image.size())
        if mean_min_conf > 0:
            with torch.no_grad():
                logits, probabilities = self.model(image, return_cls=True)
                logits = logits.cpu().squeeze(1).numpy()
                probabilities = probabilities.cpu().squeeze(1)
            conf = torch.mean(
                torch.sort(torch.max(probabilities[torch.argmax(probabilities, dim=1).to(torch.bool)], dim=1)[0])[0][
                :mean_min_conf]).item()
        else:
            with torch.no_grad():
                logits = self.model(image).cpu().squeeze(1).numpy()
        sim_pred = CTCDecoder.ctc_decode(logits, decoder_name=self.decode_method, label2char=self.label2char)
        if get_string:
            sim_pred = "".join(sim_pred)
        if mean_min_conf > 0:
            return sim_pred, conf
        else:
            return sim_pred

    def infer_group(self, images: Union[List[np.ndarray]]):
        images = TorchVisionUtils.transform_concatenate_images(images, self.transformer, device=self.device)
        with torch.no_grad():
            preds = self.model(images).squeeze(0).numpy()
        sim_preds = CTCDecoder.ctc_decode_batch(preds, decoder_name=self.decode_method, label2char=self.label2char)
        return sim_preds

    @staticmethod
    def get_stats(cls_logits):
        probs = torch.max(cls_logits, 2)[0].T.detach().numpy()
        min, max, mean, std = np.min(probs, 1), np.max(probs, 1), np.mean(probs, 1), np.std(probs, 1)
        return dict(min=min.tolist(), max=max.tolist(), mean=mean.tolist(), std=std.tolist())

    def get_reliability(self, stats):
        reliables = []
        for i in range(len(stats['min'])):
            min = stats['min'][i]
            std = stats['std'][i]
            mean = stats['mean'][i]
            if min < self.min_t or std > self.std_t or mean < self.mean_t:
                reliables.append(False)
            else:
                reliables.append(True)
        return reliables



MATPLOTLIB_COLORS = ["green", "orange", "blue", "black"]


def show_plt(img, is_rgb, figsize=None):
    import matplotlib

    matplotlib.use("TkAgg")
    import matplotlib.pyplot as plt

    if not is_rgb:
        img = img[..., ::-1]
    plt.figure(figsize=figsize)
    plt.imshow(img)
    plt.show()



import sys
import numpy as np
from deep_utils.utils.box_utils.boxes import Box
from deep_utils.utils.lib_utils.download_utils import download_decorator
from deep_utils.utils.lib_utils.lib_decorators import (
    expand_input,
    get_elapsed_time,
    get_from_config,
    rgb2bgr,
)
from deep_utils.vision.face_detection.main.main_face_detection import FaceDetector

from .config import Config
from .utils.fd_config import define_img_size
from .utils.mb_tiny_fd import create_mb_tiny_fd, create_mb_tiny_fd_predictor
from .utils.mb_tiny_RFB_fd import create_Mb_Tiny_RFB_fd, create_Mb_Tiny_RFB_fd_predictor

define_img_size(320)


class UltralightTorchFaceDetector(FaceDetector):
    def __init__(self, **kwargs):
        super().__init__(
            name=self.__class__.__name__,
            file_path=__file__,
            download_variables=("model",),
            **kwargs
        )
        self.config: Config

    @download_decorator
    def load_model(self):
        # LOAD MODELS
        model_path = self.config.model
        if self.config.model_name == "slim":
            net = create_mb_tiny_fd(is_test=True, device=self.config.device)
            net.load(model_path)
            predictor = create_mb_tiny_fd_predictor(net, device=self.config.device)
        elif self.config.model_name == "rbf":
            net = create_Mb_Tiny_RFB_fd(is_test=True, device=self.config.device)
            net.load(model_path)
            predictor = create_Mb_Tiny_RFB_fd_predictor(net, device=self.config.device)
        else:
            print("The net type is wrong!")
            sys.exit(1)
        self.model = predictor

    @get_from_config
    @get_elapsed_time
    @rgb2bgr("rgb")
    @expand_input(3)
    def detect_faces(
            self,
            images,
            is_rgb,
            confidence=None,
            get_time=False,
    ):

        if images.ndim < 4:
            images = np.array([images])

        boxes_, confidences_, landmarks_ = [], [], []
        for img_n in range(images.shape[0]):
            bboxes, labels, probs = self.model.predict(images[img_n], 500, self.config.confidence)
            conf, box = [], []
            for i in range(bboxes.size(0)):
                box_tensor = bboxes[i, :]
                boxes = [
                    box_tensor[0].item(),
                    box_tensor[1].item(),
                    box_tensor[2].item(),
                    box_tensor[3].item(),
                ]
                boxes = Box.box2box(
                    boxes, in_source=Box.BoxSource.Torch, to_source=Box.BoxSource.Numpy
                )
                confidence_ = probs[i]
                if confidence_ >= self.config.confidence:
                    conf.append(confidence_.item())
                    box.append(boxes)
            confidences_.append(conf)
            boxes_.append(box)
        return dict(boxes=boxes_, confidences=confidences_, landmarks=landmarks_)



class Colors:
    """
    Taken from ultralytics
    # Ultralytics color palette https://ultralytics.com/
    """
    def __init__(self):
        # hex = matplotlib.colors.TABLEAU_COLORS.values()
        hex = ('FF3838', 'FF9D97', 'FF701F', 'FFB21D', 'CFD231', '48F90A', '92CC17', '3DDB86', '1A9334', '00D4BB',
               '2C99A8', '00C2FF', '344593', '6473FF', '0018EC', '8438FF', '520085', 'CB38FF', 'FF95C8', 'FF37C7')
        self.palette = [self.hex2rgb('#' + c) for c in hex]
        self.n = len(self.palette)

    def __call__(self, i, bgr=False):
        c = self.palette[int(i) % self.n]
        return (c[2], c[1], c[0]) if bgr else c

    @staticmethod
    def hex2rgb(h):  # rgb order (PIL)
        return tuple(int(h[1 + i:1 + i + 2], 16) for i in (0, 2, 4))



from deep_utils.utils.opencv_utils.main import show_destroy_cv2


class AugmentTorch:
    @staticmethod
    def random_resized_crop(
            size, scale=(0.08, 1.0), ratio=(3.0 / 4.0, 4.0 / 3.0), interpolation="BILINEAR"
    ):
        from torchvision import transforms

        if interpolation == "BILINEAR":
            from torchvision.transforms import InterpolationMode

            interpolation = InterpolationMode.BILINEAR
        return transforms.RandomResizedCrop(
            size, scale=scale, ratio=ratio, interpolation=interpolation
        )

    @staticmethod
    def five_crop(size):
        from torchvision import transforms

        return transforms.FiveCrop(size)

    @staticmethod
    def ten_crop(size):
        from torchvision import transforms

        return transforms.TenCrop(size)

    @staticmethod
    def random_horizontal_flip(p=0.5):
        from torchvision import transforms

        return transforms.RandomHorizontalFlip(p=p)

    @staticmethod
    def random_affine(
            degrees,
            translate=None,
            scale=None,
            shear=None,
            interpolation="NEAREST",
            fill=0,
            fillcolor=None,
            resample=None,
    ):
        from torchvision import transforms

        if interpolation == "NEAREST":
            from torchvision.transforms import InterpolationMode

            interpolation = InterpolationMode.NEAREST
        return transforms.RandomAffine(
            degrees,
            translate=translate,
            scale=scale,
            shear=shear,
            interpolation=interpolation,
            fill=fill,
            fillcolor=fillcolor,
            resample=resample,
        )

    @staticmethod
    def random_rotation(
            degrees,
            interpolation="NEAREST",
            expand=False,
            center=None,
            fill=0,
            resample=None,
    ):
        from torchvision import transforms

        if interpolation == "NEAREST":
            from torchvision.transforms import InterpolationMode

            interpolation = InterpolationMode.NEAREST
        return transforms.RandomRotation(
            degrees,
            interpolation=interpolation,
            expand=expand,
            center=center,
            fill=fill,
            resample=resample,
        )

    @staticmethod
    def gaussian_blur(kernel_size, sigma=(0.1, 2.0)):
        from torchvision import transforms

        return transforms.GaussianBlur(kernel_size=kernel_size, sigma=sigma)

    @staticmethod
    def random_erasing(
            p=0.5, scale=(0.02, 0.33), ratio=(0.3, 3.3), value=0, inplace=False
    ):
        from torchvision import transforms

        return transforms.RandomErasing(
            p=p, scale=scale, ratio=ratio, value=value, inplace=inplace
        )

    @staticmethod
    def center_crop(size):
        from torchvision import transforms

        return transforms.CenterCrop(size)

    @staticmethod
    def color_jitter(brightness=0, contrast=0, saturation=0, hue=0):
        from torchvision import transforms

        return transforms.ColorJitter(
            brightness=brightness, contrast=contrast, saturation=saturation, hue=hue
        )

    @staticmethod
    def random_vertical_flip(p=0.5):
        from torchvision import transforms

        return transforms.RandomVerticalFlip(p=p)

    @staticmethod
    def pad(padding, fill=0, padding_mode="constant"):
        from torchvision import transforms

        return transforms.Pad(padding=padding, fill=fill, padding_mode=padding_mode)

    @staticmethod
    def random_crop(
            size, padding=None, pad_if_needed=False, fill=0, padding_mode="constant"
    ):
        from torchvision import transforms

        return transforms.RandomCrop(
            size=size,
            padding=padding,
            pad_if_needed=pad_if_needed,
            fill=fill,
            padding_mode=padding_mode,
        )

    @staticmethod
    def random_perspective(
            distortion_scale=0.5, p=0.5, interpolation="BILINEAR", fill=0
    ):
        from torchvision import transforms

        if interpolation == "BILINEAR":
            from torchvision.transforms import InterpolationMode

            interpolation = InterpolationMode.BILINEAR
        return transforms.RandomPerspective(
            distortion_scale=distortion_scale,
            p=p,
            interpolation=interpolation,
            fill=fill,
        )

    @staticmethod
    def resize(size, interpolation="BILINEAR"):
        from torchvision import transforms

        if interpolation == "BILINEAR":
            from torchvision.transforms import InterpolationMode

            interpolation = InterpolationMode.BILINEAR
        return transforms.Resize(size=size, interpolation=interpolation)

    @staticmethod
    def normalize(mean, std, inplace=False):
        from torchvision import transforms

        return transforms.Normalize(mean=mean, std=std, inplace=inplace)

    @staticmethod
    def to_tensor():
        from torchvision import transforms

        return transforms.ToTensor()

    @staticmethod
    def get_augments(*args):
        import torch
        from torchvision import transforms

        transformations = []
        use_lambda = False
        for arg in args:
            prob = 0
            arg_type = type(arg)
            if arg_type is tuple:
                arg, prob = arg
            if type(arg) is callable:
                transform = arg()
            else:
                transform = arg
            if prob:
                transform = transforms.RandomApply([transform], p=prob)
            if isinstance(transform, (transforms.TenCrop, transforms.FiveCrop)):
                use_lambda = True
            if isinstance(transform, transforms.Normalize):
                if use_lambda:
                    to_append = transforms.Lambda(
                        lambda tensors: torch.stack(
                            [AugmentTorch.to_tensor()(t) for t in tensors]
                        )
                    )
                else:
                    to_append = AugmentTorch.to_tensor()
                transformations.append(to_append)
            if use_lambda:
                transform = transforms.Lambda(
                    lambda tensors: torch.stack(
                        [transform(t) for t in tensors])
                )
            transformations.append(transform)
        return transforms.Compose(transformations)


def tensor_to_image(tensor, mean=None, std=None, return_array=True):
    """
    Note: Whenever the std is 255 the code returns wrong answers because converting to PIL format by default convert the
     image to type uint8 which has 255 in it as well.
    :param tensor:
    :param mean:
    :param std:
    :param return_array:
    :return:
    """
    import numpy as np
    import torch
    from torchvision import transforms

    if mean is not None or std is not None:
        c = tensor.size()[0]
        if std is None:
            std = torch.ones(c)
        elif mean is None:
            mean = torch.zeros(c)

        tensor = transforms.Normalize(
            mean=[-m / s for m, s in zip(mean, std)], std=[1 / s for s in std]
        )(tensor)

    image = transforms.ToPILImage()(tensor)
    if return_array:
        image = np.array(image)
    return image


def visualize_data_loader(data_loader, mean, std):
    """
    This function is used to convert tensor images to numpy images that will be shown by opencv!
    :param data_loader:
    :param mean:
    :param std:
    :return:
    """
    for x_data, *_ in data_loader:
        for image_tensor in x_data:
            cv2_image = tensor_to_image(image_tensor, mean, std)
            show_destroy_cv2(cv2_image)



from os.path import join
from transformers import AutoFeatureExtractor, AutoModelForAudioClassification
from typing import Union
from pathlib import Path
import torch
import torchaudio
import librosa
import numpy as np
from deep_utils.utils.pickle_utils.pickles import load_pickle
from deep_utils.utils.logging_utils.logging_utils import log_print
from deep_utils.audio.audio_utils.torchaudio_utils import TorchAudioUtils


class AudioClassificationWav2vec2Torch:
    def __init__(self, model_path, num_labels, device="cpu", sr=16_000, label2id="label2id.pkl",
                 feature_extractor="facebook/wav2vec2-base", logger=None, verbose=1):
        """

        :param num_labels:
        :param model_path: A path to a directory which contains pytorch_model.bin, training_args.bin, and label2id variable
        :param feature_extractor:
        """
        self.feature_extractor = AutoFeatureExtractor.from_pretrained(feature_extractor)
        self.label2id = load_pickle(join(model_path, label2id))
        self.id2label = {v: k for k, v in self.label2id.items()}
        self.sr = sr
        self.model = AutoModelForAudioClassification.from_pretrained(
            model_path, num_labels=num_labels, label2id=self.label2id, id2label=self.id2label
        ).to(device)
        self.model = self.model.eval()
        self.device = device
        self.logger = logger
        self.verbose = verbose
        log_print(self.logger, "Successfully created Wav2Vec2 Classification Object")

    def infer(self, wave: Union[Path, str, np.ndarray, torch.Tensor], sr=None):
        if isinstance(wave, (Path, str)):
            waveform, sr = librosa.load(wave)
            waveform = torch.from_numpy(waveform).unsqueeze(0)
        else:
            waveform = wave
        waveform = TorchAudioUtils.resample(waveform, sr, self.sr)
        # waveform = torchaudio.transforms.Resample(sr, self.sr)(waveform)
        inputs = self.feature_extractor(waveform, sampling_rate=self.feature_extractor.sampling_rate,
                                        max_length=16000, truncation=True)
        tensor = torch.tensor(inputs['input_values'][0]).to(self.device)
        with torch.no_grad():
            output = self.model(tensor)
            logits = output['logits'][0]
            label_id = torch.argmax(logits).item()
        label_name = self.id2label[str(label_id)]
        return label_name

    def infer_group(self, audio_segments, sr, indices=None):
        suffix = []
        indices = ["" for _ in range(len(audio_segments))] if indices is None else indices
        delimiter = '_' if indices is not None else ''
        for index, audio_segment in zip(indices, audio_segments):
            gender = self.infer(audio_segment, sr=sr)
            suffix.append(str(index) + f"{delimiter}{gender}")
        log_print(self.logger, f"Successfully performed {len(suffix)} gender detection", verbose=self.verbose)
        return suffix



import os
from abc import abstractmethod
from deep_utils.main_abs.main import MainClass
from deep_utils.utils.dict_named_tuple_utils import dictnamedtuple
from deep_utils.utils.dir_utils.dir_utils import remove_create
from deep_utils.utils.box_utils.boxes import Box
from deep_utils.utils.os_utils.os_path import split_extension
from deep_utils.utils.logging_utils.logging_utils import log_print

OUTPUT_CLASS = dictnamedtuple(
    "FaceDetector", ["boxes", "landmarks", "confidences"])


class FaceDetector(MainClass):
    def __init__(self, name, file_path, **kwargs):
        super().__init__(name, file_path=file_path, **kwargs)
        self.output_class = OUTPUT_CLASS

    @abstractmethod
    def detect_faces(self, img, is_rgb, confidence=None, get_time=False) -> OUTPUT_CLASS:
        pass

    def detect_crop_dir(
            self,
            image_directory,
            confidence=None,
            extensions=(".png", ".jpg", ".jpeg"),
            res_dir=None,
            remove_res_dir=False,
            get_biggest=False
    ):
        import cv2
        results = dict()
        remove_create(res_dir, remove=remove_res_dir)
        for item_name in os.listdir(image_directory):
            _, extension = os.path.splitext(item_name)
            if extension in extensions:
                img_path = os.path.join(image_directory, item_name)
                img = cv2.imread(img_path)
                result = self.detect_faces(
                    img,
                    is_rgb=False,
                    confidence=confidence,
                    get_time=True,
                )
                print(
                    f'{img_path}: objects= {len(result["boxes"])}, time= {result["elapsed_time"]}'
                )
                boxes = result["boxes"]
                if get_biggest:
                    boxes = [Box.get_biggest(boxes)]
                cropped_images = Box.get_box_img(img, boxes)
                result = dict(result)
                result['cropped_images'] = cropped_images
                if res_dir:
                    for e, cropped_image in enumerate(cropped_images):
                        cv2.imwrite(os.path.join(res_dir, split_extension(item_name, suffix=f"_{e}")), cropped_image)
                results[img_path] = result
        return results

    def detect_crop_dir_of_dir(
            self,
            input_directory,
            image_dir_name="images",
            cropped_dir_name="cropped",
            confidence=None,
            extensions=(".png", ".jpg", ".jpeg"),
            remove_cropped=True,
            get_biggest=False
    ):
        for directory_name in sorted(os.listdir(input_directory)):

            directory_path = os.path.join(input_directory, directory_name)
            images_dir = os.path.join(directory_path, image_dir_name)
            cropped_dir = os.path.join(directory_path, cropped_dir_name)
            if not os.path.isdir(directory_path) or not os.path.isdir(images_dir):
                log_print(None, f"Skip {directory_path}...")
                continue
            remove_create(cropped_dir, remove=remove_cropped)
            self.detect_crop_dir(images_dir, confidence=confidence, extensions=extensions, res_dir=cropped_dir,
                                 remove_res_dir=remove_cropped, get_biggest=get_biggest)



import os
import numpy as np
from abc import abstractmethod
from deep_utils.utils.pickle_utils.pickles import dump_pickle
from deep_utils.main_abs.main import MainClass
from deep_utils.utils.dict_named_tuple_utils import dictnamedtuple
from deep_utils.utils.dir_utils.dir_utils import remove_create
from deep_utils.utils.os_utils.os_path import split_extension
from deep_utils.utils.logging_utils.logging_utils import log_print

OUTPUT_CLASS = dictnamedtuple(
    "FaceRecognizer", ["encodings"])


class FaceRecognition(MainClass):
    def __init__(self, name, file_path, **kwargs):
        super().__init__(name, file_path=file_path, **kwargs)
        self.output_class = OUTPUT_CLASS
        self.normalizer = self.load_normalizer(self.config.normalizer)

    @abstractmethod
    def extract_faces(self, img, is_rgb, get_time=False) -> OUTPUT_CLASS:
        pass

    def extract_dir(
            self,
            image_directory,
            extensions=(".png", ".jpg", ".jpeg"),
            res_dir=None,
            remove_res_dir=False,
            get_mean=False,
    ):
        import cv2
        results = dict()
        remove_create(res_dir, remove=remove_res_dir)
        for item_name in os.listdir(image_directory):
            _, extension = os.path.splitext(item_name)
            if extension in extensions:
                img_path = os.path.join(image_directory, item_name)
                img = cv2.imread(img_path)
                result = self.extract_faces(img, is_rgb=False, get_time=True, )
                print(f'{img_path}: time= {result["elapsed_time"]}')

                if res_dir and not get_mean:
                    dump_pickle(os.path.join(res_dir, split_extension(item_name, extension=".pkl")), result.encodings)
                results[img_path] = result['encodings']
        if get_mean:

            encode = np.sum(np.array(list(results.values())), axis=0)
            encode = self.normalizer.transform(np.expand_dims(encode, axis=0))[0]
            results['mean-encoding'] = encode
            dump_pickle(os.path.join(res_dir, "mean-encoding.pkl"), encode)
        return results

    def extract_dir_of_dir(
            self,
            input_directory,
            image_dir_name="cropped",
            encoding_dir_name="encodings",
            extensions=(".png", ".jpg", ".jpeg"),
            remove_encoding=True,
            get_mean=False,
    ):
        results = dict()
        for directory_name in sorted(os.listdir(input_directory)):
            directory_path = os.path.join(input_directory, directory_name)
            images_dir = os.path.join(directory_path, image_dir_name)
            cropped_dir = os.path.join(directory_path, encoding_dir_name)
            if not os.path.isdir(directory_path) or not os.path.isdir(images_dir):
                log_print(None, f"Skip {directory_path}...")
                continue
            remove_create(cropped_dir, remove=remove_encoding)
            dir_result = self.extract_dir(images_dir, extensions=extensions, res_dir=cropped_dir,
                                          remove_res_dir=remove_encoding,
                                          get_mean=get_mean)
            results[directory_name] = dir_result
        if get_mean:
            results = {k: v['mean-encoding'] for k, v in results.items()}
            encoding_name = os.path.split(input_directory)[-1]
            dump_pickle(os.path.join(input_directory, encoding_name + ".pkl"), results)
        return results

    @staticmethod
    def load_normalizer(normalizer_name):

        if normalizer_name == "l2_normalizer":
            from sklearn.preprocessing import Normalizer
            l2_normalizer = Normalizer('l2')
            return l2_normalizer
        return lambda x: x


from collections import OrderedDict
from typing import Union

import torch
from torch import nn

from deep_utils.utils.list_utils.list_utils import shift_lst


class BlocksTorch:
    @staticmethod
    def conv_norm_act(
            in_c,
            out_c,
            k=(3, 3),
            s=(1, 1),
            p=(1, 1),
            norm=None,
            act="relu",
            conv=True,
            index=None,
            conv_kwargs: dict = None,
            norm_kwargs: dict = None,
            act_kwargs: dict = None,
            pooling=None,
            pooling_k=(2, 2),
            pooling_s=(2, 2),
            pool_kwargs=None,
            padding: Union[None, str] = None,
            move_forward=0,
    ):
        """

        :param in_c:
        :param out_c:
        :param k:
        :param s:
        :param p:
        :param norm:
        :param act:
        :param conv:
        :param index:
        :param conv_kwargs:
        :param norm_kwargs:
        :param act_kwargs:
        :param pooling:
        :param pooling_k:
        :param pooling_s:
        :param pool_kwargs:
        :param padding:
        :param move_forward: Move forward will shift the conv and act and bn, consider a block should start with a relu not a cnn...
        :return:
        """
        conv_kwargs = dict() if conv_kwargs is None else conv_kwargs
        norm_kwargs = dict() if norm_kwargs is None else norm_kwargs
        act_kwargs = dict() if act_kwargs is None else act_kwargs
        pool_kwargs = dict() if pool_kwargs is None else pool_kwargs
        k = (k, k) if isinstance(k, int) else k
        s = (s, s) if isinstance(s, int) else s
        p = (p, p) if isinstance(p, int) else p
        pooling_s = (pooling_s, pooling_s) if isinstance(
            pooling_s, int) else pooling_s
        pooling_k = (pooling_k, pooling_k) if isinstance(
            pooling_k, int) else pooling_k
        modules, names = [], []
        if conv:
            if padding and padding.lower() == "same":
                p = (k[0] // (2 * s[0]), k[1] // (2 * s[1]))
            modules.append(BlocksTorch.conv_2d(
                in_c, out_c, k, s, p, **conv_kwargs))
            names.append(f"conv" + (f"_{index}" if index else ""))
        if norm:
            modules.append(
                BlocksTorch.load_layer_norm(norm, out_c=out_c, **norm_kwargs)
            )
            names.append(f"{norm}" + (f"_{index}" if index else ""))
        if act:
            if act == "relu":
                act_kwargs_ = {"inplace": True}
                act_kwargs_.update(act_kwargs)
                act_kwargs = act_kwargs_
            modules.append(BlocksTorch.load_activation(act, **act_kwargs))
            names.append(f"{act}" + (f"_{index}" if index else ""))
        modules = shift_lst(modules, move_forward=move_forward % 3)
        names = shift_lst(names, move_forward=move_forward % 3)
        if pooling:
            pooling_layer = BlocksTorch.load_pooling(
                pooling, pooling_k, pooling_s, **pool_kwargs
            )
            modules.append(pooling_layer)
            names.append(f"{pooling}" + (f"_{index}" if index else ""))
        cnn = nn.Sequential(
            OrderedDict({name: module for name, module in zip(names, modules)})
        )
        return cnn

    @staticmethod
    def load_pooling(pooling, k, s, **pooling_kwargs):
        if pooling == "max":
            pooling = nn.MaxPool2d(k, s, **pooling_kwargs)
        elif pooling == "avg":
            pooling = nn.AvgPool2d(k, s, **pooling_kwargs)
        else:
            raise ValueError(
                f"requested pooling {pooling} is not supported. Supported poolings: max, avg"
            )
        return pooling

    @staticmethod
    def load_activation(act, **act_kwargs):
        if act == "relu":
            activation = nn.ReLU(**act_kwargs)
        elif act == "leaky_relu":
            activation = nn.LeakyReLU(**act_kwargs)
        else:
            raise ValueError(
                f"requested activation {act} is not supported. Supported activations: relu, leaky_relu"
            )
        return activation

    @staticmethod
    def conv_2d(
            in_c,
            out_c,
            k=(3, 3),
            s=(1, 1),
            p=(1, 1),
            groups=1,
            bias=True,
            padding_mode="zeros",
    ):
        import torch

        return torch.nn.Conv2d(
            in_c, out_c, k, s, p, groups=groups, bias=bias, padding_mode=padding_mode
        )

    @staticmethod
    def load_layer_norm(norm, out_c, **norm_kwargs):
        if norm == "bn":
            layer_norm = nn.BatchNorm2d(num_features=out_c, **norm_kwargs)
        elif norm == "gn":
            layer_norm = nn.GroupNorm(num_channels=out_c, **norm_kwargs)
        else:
            raise ValueError(
                f"requested layer_norm {norm} is not supported. Supported layer_norms: bn(batch normalization), "
                f"gn(group normalization"
            )
        return layer_norm

    @staticmethod
    def res_basic_block(in_c, out_c, bias, down_sample=False):
        modules = []
        names = []
        for i in range(2):
            modules.append(
                BlocksTorch.conv_norm_act(
                    int(in_c / 2) if down_sample and i == 0 else in_c,
                    out_c,
                    act=False if i == 1 else "relu",
                    s=(2, 2) if down_sample and i == 0 else (1, 1),
                    conv_kwargs={"bias": bias},
                )
            )
            names.append(f"cnn_{i}_in{in_c}_f{out_c}")
        res_block = nn.Sequential(
            OrderedDict({name: module for name, module in zip(names, modules)})
        )
        return res_block

    @staticmethod
    def fc_dropout(
            in_f,
            out_f,
            act="relu",
            drop_p=0.0,
            fc=True,
            index=None,
            fc_kwargs: Union[None, dict] = None,
            act_kwargs: Union[None, dict] = None,
            drop_kwargs: Union[dict, None] = None,
    ):
        fc_kwargs = dict() if fc_kwargs is None else fc_kwargs
        drop_kwargs = dict() if drop_kwargs is None else drop_kwargs
        act_kwargs = dict() if act_kwargs is None else act_kwargs

        modules, names = [], []
        if fc:
            modules.append(nn.Linear(in_f, out_f, **fc_kwargs))
            names.append(f"fc" + (f"_{index}" if index else ""))
        if act:
            if act == "relu":
                act_kwargs_ = {"inplace": True}
                act_kwargs_.update(act_kwargs)
                act_kwargs = act_kwargs_
            modules.append(BlocksTorch.load_activation(act, **act_kwargs))
            names.append(f"{act}" + (f"_{index}" if index else ""))
        if drop_p:
            modules.append(nn.Dropout(drop_p, **drop_kwargs))
            names.append(f"dropout" + (f"_{index}" if index else ""))
        return nn.Sequential(
            OrderedDict({name: module for name, module in zip(names, modules)})
        )

    @staticmethod
    def weights_init(m):
        class_name = m.__class__.__name__
        if class_name.find("Conv") != -1:
            m.weight.data.normal_(0.0, 0.02)
        elif class_name.find("BatchNorm") != -1:
            m.weight.data.normal_(1.0, 0.02)
            m.bias.data.fill_(0)
        elif class_name.find("Linear") != -1:
            torch.nn.init.xavier_uniform(m.weight)
            m.bias.data.fill_(0.01)
        elif class_name == "LSTM":
            # setting bias to zero and weight with xavier
            for name, param in m.named_parameters():
                if "bias" in name:
                    nn.init.constant(param, 0.0)
                elif "weight" in name:
                    nn.init.xavier_normal(param)

    @staticmethod
    def set_parameter_requires_grad(model: nn.Module, feature_extracting):
        if feature_extracting:
            for param in model.parameters():
                param.requires_grad = False

    @staticmethod
    def set_parameter_grad(model: nn.Module, num_layer):
        for param in list(model.parameters())[::-1][:num_layer]:
            param.requires_grad = True


if __name__ == "__main__":
    a = BlocksTorch.conv_norm_act(32, 64, move_forward=2)
    print(a)
    b = BlocksTorch.fc_dropout(64, 124, drop_p=0.5)
    print(b)



from pathlib import Path
from typing import Union

from deep_utils.utils.logging_utils import log_print
from deep_utils.utils.utils.yaml_utils import dump_yaml, load_yaml, yaml_post_process


class KeyValStruct:
    """
    This is a simple class for hyperparameter tuning
    """

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            if isinstance(v, dict):
                self.__dict__[k] = KeyValStruct(**v)
            else:
                self.__dict__[k] = v


class YamlConfig(KeyValStruct):
    """
    A simple yaml config generator!
    """

    @classmethod
    def load_config(cls, config_path: str):
        config_dict = load_yaml(config_path)
        config_dict = yaml_post_process(config_dict)
        return cls(**config_dict)


def keyval_struct2dict(key_val_struct: KeyValStruct):
    """
    Converts a keyval structure to dictionary
    Args:
        key_val_struct:

    Returns:

    """
    output = dict()
    for k, v in vars(key_val_struct).items():
        if isinstance(v, KeyValStruct):
            output[k] = keyval_struct2dict(v)
        else:
            output[k] = v
    return output


def yaml_config2yaml_file(
    yaml_config: YamlConfig, yaml_path: Union[str, Path], logger=None, verbose=1
):
    """
    Dumps aa yaml config object to a yaml file
    Args:
        yaml_config:
        yaml_path:
        logger:
        verbose:

    Returns:

    """
    dict_obj = keyval_struct2dict(yaml_config)
    dump_yaml(dict_obj, yaml_path, logger=logger, verbose=verbose)
    log_print(
        logger, f"Successfully saved yaml_config in {yaml_path}", verbose=verbose)



def get_counter_name(name:str, name_dict: dict, counter=0, separator="_"):
    """
    increase the counter till it reaches a number that is not in the name_dict
    :param name:
    :param name_dict:
    :param counter:
    :param separator:
    :return:
    """
    while name + f"{separator}{counter}" in name_dict:
        counter += 1
    return name + f"{separator}{counter}"



import random
from collections import defaultdict
from copy import deepcopy
from typing import Callable, Dict, List, Tuple, Union

from deep_utils.nlp.ner.base_operations import NEROperation
from deep_utils.nlp.ner.taggers import bio_tagger
from deep_utils.nlp.ner.utils_ import check_bio_labels
from deep_utils.utils.random_utils.random_utils import random_choice_group


class NERAugRemove(NEROperation):
    """
    This augmentation randomly removes tokens and their corresponding labels, This is not just for persian language.
    >>> remove_dict = { ("-", "o"): 0.5, (",", "o"): 0.5, ("استان", "province"): 0.5, ("شهر", "city"): 0.5}

    >>> seed = 1234
    >>> remove_aug = NERAugRemove(remove_dict,random_seed=seed)
    >>> tokens = ['استان', 'تهران', '-', 'شهر', 'تهران', '-', 'نعمت\u200cآباد', '-', 'خیابان', 'نور', '-', 'کوچه', '21', '-', 'پلاک', '23', '-', 'تلفن', '982155871442']
    >>> labels = ['B-province',  'I-province', 'o', 'B-city', 'I-city', 'o', 'B-section', 'o', 'B-street', 'I-street', 'o', 'B-alley', 'I-alley', 'o', 'B-house_number', 'I-house_number', 'o', 'B-phone', 'I-phone']
    >>> remove_aug(tokens, labels)
    (['استان', 'تهران', '-', 'تهران', 'نعمت\u200cآباد', 'خیابان', 'نور', '-', 'کوچه', '21', '-', 'پلاک', '23', '-', 'تلفن', '982155871442'], ['B-province', 'I-province', 'o', 'B-city', 'B-section', 'B-street', 'I-street', 'o', 'B-alley', 'I-alley', 'o', 'B-house_number', 'I-house_number', 'o', 'B-phone', 'I-phone'])
    """

    def __init__(
        self,
        remove_dict: Dict[Tuple[str, str], float],
        tokenizer: Callable[[str], List[str]] = str.split,
        ner_type="BIO",
        make_copy=True,
        random_seed=None,
    ):
        """

        :param remove_dict:
        :param tokenizer:
        :param ner_type:
        :param make_copy: whether to create copy of the input tokens and labels to prevent data modification, the default is True
        :param random_seed:
        """
        super(NERAugRemove, self).__init__(ner_type=ner_type)
        if random_seed:
            random.seed(random_seed)
        self.tokenizer = tokenizer
        self.make_copy = make_copy
        self.p_dict = {}
        self.augmentation_len_dict = defaultdict(list)
        for (query_tokens, query_label), prob in remove_dict.items():
            query_tokens = tuple(self.tokenizer(query_tokens))
            key = (query_tokens, query_label)
            # Getting probabilities from the vals
            self.p_dict[key] = prob
            # This code will convert str to list[str] if the input labels were string!
            # Getting the labels from vals
            self.augmentation_len_dict[len(query_tokens)].append(
                (query_tokens, query_label)
            )
        self.aug_tokens_lengths = sorted(
            list(self.augmentation_len_dict.keys()), reverse=True
        )

    def __call__(self, tokens_x: List[str], labels_y: List[str]):
        if self.make_copy:
            tokens_x, labels_y = deepcopy(tokens_x), deepcopy(labels_y)
        super().__call__(tokens_x, labels_y)

        for token_len in self.aug_tokens_lengths:
            for query_tokens, query_label in self.augmentation_len_dict[token_len]:
                token_index = 0
                p = self.p_dict[(query_tokens, query_label)]
                while token_index <= (len(tokens_x) - token_len):
                    selected_tokens = tokens_x[token_index: token_index + token_len]
                    selected_labels = labels_y[token_index: token_index + token_len]
                    if (
                        query_tokens == tuple(selected_tokens)
                        and (
                            not query_label
                            or check_bio_labels(selected_labels, query_label)
                        )
                        and random.random() <= p
                    ):
                        remove_label = labels_y[token_index: token_index + token_len][
                            0
                        ]
                        if remove_label.startswith("B"):
                            # If it starts with B it means by removing this one the others fragments won't have Beginning
                            # tag, so the next tag should be changed to start with "B" instead of I
                            next_index = token_index + token_len
                            if (
                                next_index < len(labels_y)
                                and labels_y[next_index].startswith("I")
                                and labels_y[next_index][2:] == query_label
                            ):
                                labels_y[next_index] = (
                                    "B-" + query_label
                                )  # start with B

                        del tokens_x[token_index: token_index + token_len]
                        del labels_y[token_index: token_index + token_len]
                        token_index -= token_len
                        # continue
                    token_index += 1

        return tokens_x, labels_y


class NERAugReplacement(NEROperation):
    """
    This class replaces a single token with a single token
    """

    def __init__(
        self,
        replacement_dict: Dict[Tuple[str, Union[None, str]], Tuple[List[str], float]],
        tokenizer: Callable[[str], List[str]] = str.split,
        ner_type="BIO",
        make_copy=True,
        random_seed=None,
    ):
        """

        :param remove_dict:
        :param tokenizer:
        :param ner_type:
        :param make_copy: whether to create copy of the input tokens and labels to prevent data modification, the default is True.
        :param random_seed:
        """
        super(NERAugReplacement, self).__init__(ner_type=ner_type)
        if random_seed:
            random.seed(random_seed)
        self.tokenizer = tokenizer
        self.make_copy = make_copy
        self.p_dict = {}
        self.augmentation_len_dict = defaultdict(list)
        for (query_tokens, query_label), (
            replacement_tokens,
            prob,
        ) in replacement_dict.items():
            query_tokens = tuple(self.tokenizer(query_tokens))
            key = (query_tokens, query_label)
            # Getting probabilities from the vals
            self.p_dict[key] = prob
            # This code will convert str to list[str] if the input labels were string!
            # Getting the labels from vals
            replacement_tokens = [
                self.tokenizer(replacement_token)
                for replacement_token in replacement_tokens
            ]
            replacement_labels = [
                bio_tagger(r_t, query_label) for r_t in replacement_tokens
            ]
            self.augmentation_len_dict[len(query_tokens)].append(
                (query_tokens, query_label, replacement_tokens, replacement_labels)
            )
        self.aug_tokens_lengths = sorted(
            list(self.augmentation_len_dict.keys()), reverse=True
        )

    def __call__(
        self,
        tokens_x: Union[Tuple[str], List[str]],
        labels_y: Union[Tuple[str], List[str]],
    ):
        if self.make_copy:
            tokens_x, labels_y = deepcopy(tokens_x), deepcopy(labels_y)
        super().__call__(tokens_x, labels_y)
        for token_len in self.aug_tokens_lengths:
            for (
                query_tokens,
                query_label,
                replace_tokens,
                replace_labels,
            ) in self.augmentation_len_dict[token_len]:
                token_index = 0
                p = self.p_dict[(query_tokens, query_label)]
                while token_index <= (len(tokens_x) - token_len):
                    selected_tokens = tokens_x[token_index: token_index + token_len]
                    selected_labels = labels_y[token_index: token_index + token_len]
                    if (
                        query_tokens == tuple(selected_tokens)
                        and (
                            not query_label
                            or check_bio_labels(selected_labels, query_label)
                        )
                        and random.random() <= p
                    ):
                        del tokens_x[token_index: token_index + token_len]
                        del labels_y[token_index: token_index + token_len]
                        replacement_token, replacement_label = random_choice_group(
                            replace_tokens, replace_labels
                        )
                        tokens_x[token_index:token_index] = replacement_token
                        labels_y[token_index:token_index] = replacement_label
                        token_index += len(replacement_token)
                        continue
                    token_index += 1

        return tokens_x, labels_y


if __name__ == "__main__":
    replacement = NERAugReplacement(
        replacement_dict={
            ("خیابان", "street"): (["خ", "خ."], 0.5),
            ("کوچه", "alley"): (["ک", "ک."], 0.5),
            ("-", "o"): ([",", "-"], 0.5),
            ("بلوار", "boulevard"): (["بل", "بل."], 0.5),
        }
    )
    removal = NERAugRemove(
        remove_dict={("-", "o"): 0.5, ("استان", "province"): 0.2})
    address = {
        "tokens": [
            "استان",
            "قم",
            "-",
            "شهر",
            "قم",
            "-",
            "-",
            "خیابان",
            "بعثت",
            "-",
            "بلوار",
            "تهرانی",
            "مقدم",
            "-",
            "پلاک",
            "146",
            "-",
            "طبقه",
            "2",
            "-",
            "تلفن",
            "989373584544",
        ],
        "labels": [
            "B-province",
            "I-province",
            "o",
            "B-city",
            "I-city",
            "o",
            "o",
            "B-street",
            "I-street",
            "o",
            "B-boulevard",
            "I-boulevard",
            "I-boulevard",
            "o",
            "B-house_number",
            "I-house_number",
            "o",
            "B-floor",
            "I-floor",
            "o",
            "B-phone",
            "I-phone",
        ],
    }
    tokens, labels = replacement(
        tokens_x=address["tokens"], labels_y=address["labels"])
    tokens, labels = removal(tokens_x=tokens, labels_y=labels)
    for t, l in zip(tokens, labels):
        print(t, l)



from .download_utils import download_decorator, download_file, unzip_func
from .lib_decorators import (
    cast_kwargs_dict,
    expand_input,
    get_elapsed_time,
    get_from_config,
    lib_rgb2bgr,
    rgb2bgr,
)
from .main_utils import import_module



import re


class REUtils:
    @staticmethod
    def cleaning(input_string: str, delimiter: str = "-"):
        cleanings = {
            "  ": " ",
            "- -": delimiter,
            "--": delimiter,
        }
        for cleaning_str, clean_val in cleanings.items():
            input_string = input_string.replace(cleaning_str, clean_val)
        return input_string.strip()

    @staticmethod
    def split_char_number_sentence(input_address):
        return " ".join(
            REUtils.split_char_number(var) for var in input_address.split(" ")
        )

    @staticmethod
    def split_word_punctuation_sentence(input_address):
        return " ".join(
            REUtils.split_word_punctuation(var) for var in input_address.split(" ")
        )

    @staticmethod
    def split_char_number(input_string, punctuations=",*!+-.،"):
        """
        creates a space between english numbers and other characters
        :param input_string:
        :param punctuations:
        :return:
         >>> REUtils.split_char_number("s12")
         's 12'
         >>> REUtils.split_char_number(",")
         ','
        """
        if "+" in punctuations:
            punctuations = punctuations.replace("+", "")
            plus = r"|\+"
        else:
            plus = ""
        return " ".join(
            re.findall(rf"[^\W\d]+|\d+|[{punctuations}]+{plus}", input_string)
        )

    @staticmethod
    def split_word_punctuation(input_string, punctuations=",*!+-.،"):
        r"""
        creates a space between characters and punctuations.
        \W is opposite of \w and ^ means it's reverse
        :param input_string:
        :param punctuations
        :return:
        >>> REUtils.split_word_punctuation("تهران+")
        'تهران +'
        >>> REUtils.split_word_punctuation("21+")
        '21 +'
        >>> REUtils.split_word_punctuation("۲۱+")
        '۲۱ +'
        >>> REUtils.split_word_punctuation("pooya+")
        'pooya +'
        """

        if "+" in punctuations:
            punctuations = punctuations.replace("+", "")
            plus = r"|\+"
        else:
            plus = ""
        return " ".join(re.findall(rf"[^\W^_]+|[{punctuations}]+{plus}", input_string))

    @staticmethod
    def replace(input_string, pattern, result):
        input_string = re.sub(pattern, result, input_string)
        return input_string

    @staticmethod
    def get_left_right_linespaces(value, left="[\s,]*", right="\s+"):
        expression = left + value + right
        return expression

    @staticmethod
    def replace_single_char(
            input_string, replace_expression, result_expression, left="[\s,]+", right="\s+"
    ):
        """
        Replaces a single character with a complete word
        :param input_string:
        :param replace_expression:
        :param result_expression:
        :param left:
        :param right:
        :return:
        """
        input_string = " " + input_string  # for those starting at index zero
        pattern = REUtils.get_left_right_linespaces(
            replace_expression, left=left, right=right
        )
        input_string = REUtils.replace(
            input_string, pattern, " " + result_expression + " "
        )
        input_string = REUtils.cleaning(
            input_string
        )  # This is done to clean up extra spaces and other characters!
        return input_string

    @staticmethod
    def remove_en_punctuations(input_string):
        """
        Removes punctuations in a string
        :param input_string:
        :return:
        >>> REUtils.remove_en_punctuations('!hi. wh?at is the weat[h]er lik?e)()()(/\.')
        'hi what is the weather like'
        """
        return re.sub(r'[^\w\s]', '', input_string)

    @staticmethod
    def replace_map(text: str, chars_to_map: dict):
        """
        This function is used to replace a dictionary of characters inside a text string
        :param text:
        :param chars_to_map:
        :return:
        """
        import re

        pattern = "|".join(map(re.escape, chars_to_map.keys()))
        return re.sub(pattern, lambda m: chars_to_map[m.group()], str(text))


if __name__ == "__main__":
    str_ = "s12"
    output = REUtils.split_char_number(str_)
    print(output)



from deep_utils.main_abs.dummy_framework.dummy_framework import DummyObject, requires_backends


class TorchVisionInference(metaclass=DummyObject):
    _backend = ["torch", "cv2", "torchvision"]

    def __init__(self, *args, **kwargs):
        requires_backends(self, self._backend)


class TorchVisionModel(metaclass=DummyObject):
    _backend = ["torch", "numpy", "torchvision"]

    def __init__(self, *args, **kwargs):
        requires_backends(self, self._backend)



import os

from deep_utils.utils.dir_utils.dir_utils import remove_create
from deep_utils.utils.logging_utils import log_print
from deep_utils.utils.json_utils import dump_json, load_json


def get_categories(json_paths):
    """
    getting categories by updating all together.
    :param json_paths:
    :return:
    """
    categories = {}
    for json_path in json_paths:
        data = load_json(json_path)
        category = data["categories"]
        categories.update(category)
        # break
    return categories


def get_img_to_name_dict(images):
    """
    Gets {img-id: img-data} dictionary
    :param images:
    :return:
    """
    dict_ = dict()
    for img_dict in images:
        id_ = img_dict.pop("id")
        dict_[id_] = img_dict
    return dict_


def getting_img_information(json_paths, img_paths):
    all_data = []
    for json_path, img_path in zip(json_paths, img_paths):
        try:
            data = load_json(json_path)
        except:
            print(f"[INFO] {json_path} has a problem, continuing...")
            continue
        img_to_name = get_img_to_name_dict(data["images"])
        for ann in data["annotations"]:
            img_id = ann["image_id"]
            img_info = img_to_name[img_id].copy()
            # Getting full address
            img_info["file_name"] = os.path.join(
                img_path, img_info["file_name"])
            ann["image_id"] = img_info
            all_data.append(ann)
        print(f"[INFO] {json_path} is done!")
    return all_data


def combine_jsons(images, file_list, des, json_name, json_paths):
    """
    combining jsons
    :param images:
    :param file_list:
    :param des:
    :param json_name:
    :param json_paths:
    :return:
    """
    import shutil
    from copy import deepcopy

    file_paths = [img_info[-1] for img_info in images]
    new = [ann for ann in file_list if ann["image_id"]
           ["file_name"] in file_paths]

    path_id = {
        file_pth: (en, width, height, str(en) +
                   "_" + os.path.basename(file_pth))
        for en, (width, height, file_pth) in enumerate(images, 1)
    }
    json_images = [
        {"id": id_, "width": width, "height": height, "file_name": file_name}
        for _, (id_, width, height, file_name) in path_id.items()
    ]
    annotations = []
    for en, ann in enumerate(deepcopy(new), 1):
        image_id = path_id[ann["image_id"]["file_name"]][0]
        ann["image_id"] = image_id
        ann["id"] = en
        annotations.append(ann)

    json_dict = {
        "info": {"description": "my-project-name"},
        "images": json_images,
        "annotations": annotations,
        "categories": get_categories(json_paths),
    }
    dump_json(json_name, json_dict)

    for file_pth, (en, width, height, name) in path_id.items():
        shutil.copy(file_pth, os.path.join(des, name))


def combine_coco_json(json_paths, image_dirs, result_img_dir, result_json_path):
    remove_create(result_img_dir)
    if os.path.isfile(result_json_path):
        os.remove(result_json_path)

    all_images = getting_img_information(json_paths, image_dirs)
    images = list(set([tuple(ann["image_id"].values()) for ann in all_images]))
    combine_jsons(images, all_images, result_img_dir,
                  result_json_path, json_paths)


def get_image_id(json_dict, img_name):
    images = json_dict["images"]
    for d in images:
        if d["file_name"] == img_name:
            return d["id"]


def get_annotation(json_dict, id_):
    annotations = json_dict["annotations"]
    annotations_list = list()
    for ann in annotations:
        if ann["image_id"] == id_:
            annotations_list.append(ann)
    return annotations_list


def get_coco_json_masks(json_path, mask_path, logger=None, verbose=1):
    """
    Converting json coco file segmentations to masks for Unet and other methods.
    """
    from collections import defaultdict

    import cv2
    import numpy as np
    from tqdm import tqdm

    remove_create(mask_path)
    json_dict = load_json(json_path)
    # get the annotations
    annotations_point = json_dict["annotations"]
    # make a dictionary from annotations with image_ids as keys for each of use.
    # img_annotation_dict = {ann['image_id']: ann for ann in annotations_point}
    img_annotation_dict = defaultdict(list)
    for ann in annotations_point:
        img_annotation_dict[ann["image_id"]].append(ann)
    # img_annotation_dict = {1: [{segmentations: ....}, {segmentations: ....}, ...]}
    # get the images' information
    # [{img_name: ... , width, id:... } , {}, ... ]
    image_file = json_dict["images"]
    # make a dictionary of images' information for ease of use.
    img_dict = {img_file["id"]: img_file for img_file in image_file}
    for img_id, img_info in tqdm(img_dict.items(), total=len(img_dict)):
        mask = np.zeros(
            (img_info["height"], img_info["width"]), dtype=np.uint8)
        annotations = img_annotation_dict[img_id]
        for ann in annotations:
            segmentations = ann["segmentation"]
            cat_id = ann["category_id"]
            for seg in segmentations:
                # getting the points
                xs = seg[0::2]
                ys = seg[1::2]
                pts = np.array([[x, y]
                               for x, y in zip(xs, ys)], dtype=np.int32)
                pts = pts.reshape((-1, 1, 2))

                # draw the points on the mask image.
                try:
                    mask = cv2.fillPoly(mask, [pts], cat_id)
                except Exception as e:
                    log_print(
                        logger,
                        f"Error for {img_info['file_name']}, len={len(pts)}",
                        verbose=verbose,
                    )
        cv2.imwrite(os.path.join(mask_path, img_info["file_name"]), mask)
        log_print(logger, f"img_id: {img_id} is done!", verbose=verbose)


def create_json(images, annotations, categories, des, json_name, main_image):
    import os
    import shutil
    from copy import deepcopy

    from tqdm import tqdm

    file_paths = [img_info["id"] for img_info in images]
    new_annotations = [
        ann for ann in annotations if ann["image_id"] in file_paths]

    old_id = dict()
    new_images = deepcopy(images)
    for en, image in enumerate(new_images, 1):
        old_id[image["id"]] = en
        image["id"] = en

    annotations = []
    for ann_id, ann in enumerate(deepcopy(new_annotations), 1):
        image_id = old_id[ann["image_id"]]
        ann["image_id"] = image_id
        ann["id"] = ann_id
        annotations.append(ann)

    json_dict = {
        "info": {"description": "my-project-name"},
        "images": new_images,
        "annotations": annotations,
        "categories": categories,
    }

    dump_json(json_name, json_dict)

    for image in tqdm(new_images, total=len(new_images)):
        file_name = image["file_name"]
        shutil.copy(os.path.join(main_image, file_name),
                    os.path.join(des, file_name))


def coco_json_train_test_split(
    main_json,
    main_image,
    train_json="./train.json",
    train_image="./train",
    val_json="./val.json",
    val_image="./val",
    split_size=0.2,
):
    from sklearn.model_selection import train_test_split

    remove_create(train_image)
    remove_create(val_image)
    if os.path.isfile(val_json):
        os.remove(val_json)
    if os.path.isfile(train_json):
        os.remove(train_json)
    main_json_dict = load_json(main_json)
    images = main_json_dict["images"]
    annotations = main_json_dict["annotations"]
    categories = main_json_dict["categories"]
    train_images, val_images = train_test_split(images, test_size=split_size)
    create_json(
        train_images, annotations, categories, train_image, train_json, main_image
    )
    create_json(val_images, annotations, categories,
                val_image, val_json, main_image)


def convert_coco_json_yolo(path_to_json, img_dir, label_path, res_img):
    import shutil

    import cv2

    data = load_json(path_to_json)
    images = data["images"]
    annotations = data["annotations"]
    for image in images:
        img_id = image["id"]
        file_name = image["file_name"]
        img_path = os.path.join(img_dir, file_name)
        img = cv2.imread(img_path)
        img_h, img_w, _ = img.shape
        bboxes = []
        for ann in annotations:
            if ann["image_id"] == img_id:
                x1, y1, w, h = ann["bbox"]
                x_c, y_c, w, h = (
                    (x1 + w / 2) / img_w,
                    (y1 + h / 2) / img_h,
                    w / img_w,
                    h / img_h,
                )
                category_id = int(ann["category_id"]) - 1
                bboxes.append([category_id, x_c, y_c, w, h])
        if bboxes:
            with open(
                os.path.join(
                    label_path, ".".join(
                        os.path.basename(file_name).split(".")[:-1])
                )
                + ".txt",
                mode="w",
            ) as f:
                for bbox_info in bboxes:
                    f.write(" ".join(str(b) for b in bbox_info))
                    f.write("\n")
            shutil.copy(img_path, res_img)



import os

from deep_utils.utils.logging_utils import log_print
from deep_utils.utils.os_utils.os_path import split_extension


def vox2wav(
    file_path,
    in_audio_type="v3",
    overwrite=True,
    wav_path: str = None,
    logger=None,
    verbose=1,
):
    if in_audio_type.lower() in ("vox", "v3"):
        wav_path = (
            split_extension(file_path, extension=".wav")
            if wav_path is None
            else wav_path
        )
        if overwrite:
            os.system(
                f'ffmpeg -f u8 -c adpcm_ima_oki -ar 6.0k -ac 1 -i "{file_path}" "{wav_path}" -y'
            )
        else:
            if os.path.isfile(wav_path):
                raise ValueError(
                    f"overwrite is set to {overwrite} while the file exists in {wav_path}"
                )
    else:
        wav_path = file_path
    log_print(
        logger, f"Successfully converted {file_path} to {wav_path}", verbose=verbose
    )
    return wav_path



try:
    from deep_utils.dummy_objects.vision.face_recognition import VggFace2TorchFaceRecognition
    from deep_utils.vision.face_recognition.vggface2.torch.vggface2_torch import VggFace2TorchFaceRecognition
except ModuleNotFoundError:
    pass



from typing import Union


def color_str(text, color, mode: Union[str, list] = "bold"):
    """
    colorful texts!
    :param text: input text
    :param color: text color
    :param mode: defines text's modes. Valid modes: [ underline, bold ]. Pass a list of modes in case more one mode is needed!
    :return: colored text
    """
    if isinstance(mode, str):
        mode = [mode]
    colors = {
        "black": "\033[30m",  # basic colors
        "red": "\033[31m",
        "green": "\033[32m",
        "yellow": "\033[33m",
        "blue": "\033[34m",
        "magenta": "\033[35m",
        "cyan": "\033[36m",
        "white": "\033[37m",
        "bright_black": "\033[90m",  # bright colors
        "bright_red": "\033[91m",
        "bright_green": "\033[92m",
        "bright_yellow": "\033[93m",
        "bright_blue": "\033[94m",
        "bright_magenta": "\033[95m",
        "bright_cyan": "\033[96m",
        "bright_white": "\033[97m",
        "end": "\033[0m",  # misc
        "bold": "\033[1m",
        "underline": "\033[4m",
    }
    return colors[color] + "".join(colors[m] for m in mode) + text + colors["end"]



try:
    from deep_utils.dummy_objects.audio.utils_ import TorchAudioUtils
    from deep_utils.audio.audio_utils.torchaudio_utils import TorchAudioUtils
except ModuleNotFoundError:
    pass

try:
    from deep_utils.dummy_objects.audio.utils_ import LibrosaUtils
    from deep_utils.audio.audio_utils.librosa_utils import LibrosaUtils
except ModuleNotFoundError:
    pass

from .vox_utils import vox2wav



from typing import Union

import numpy as np

from deep_utils.utils.algorithm_utils.main import subset_sum


def stratify_train_test_split_multi_label(
    x: Union[list, tuple, np.ndarray], y: np.ndarray, test_size=0.2, closest_ratio=False
):
    """
        A handy function for splitting multi-label samples based on their number of classes. This is mainly useful for
    object detection and ner-like tasks that each sample may contain several objects/tags from different classes! The
    process of splitting starts from classes with the smallest number of samples to make sure their ratio is saved
    because they have small numbers of samples, retaining the ratio for them is challenging compared to those classes
    with more samples
    :param x: A list, Tuple or ndarray that contains the samples
    :param y: A 2D array that represents the number of labels in each class. Each column is representative of a class.
    As an example: y = np.array([[2, 3], [1, 1]]) says that sample one has
    two objects/tags for class 0 and 3 objects/tags for class 1 and so on
    :param test_size: size of the test set
    :param closest_ratio: For huge arrays extracting the closest ratio requires an intensive recursive function to work
     which could result in maximum recursion error. Being set to True will choose samples from the those with the smallest difference to the target number to ensure the best ratio. Set this variable to True if you are sure. by default is set to False.
    :return:
    >>> y = np.array([[1, 2, 0], [1, 0, 0], [1, 2, 0]])
    >>> x = np.array([[1, 1, 1], [2, 2, 2], [3, 3, 3]])
    >>> stratify_train_test_split_multi_label(x, y, test_size=0.3)
    (array([[2, 2, 2],
           [3, 3, 3]]), array([[1, 1, 1]]), array([[1, 0, 0],
           [1, 2, 0]]), array([[1, 2, 0]]))
    >>> x = np.array([[1, 2], [3, 4], [1, 2], [3, 4], [1, 2], [3, 4], [1, 2], [3, 4]])
    >>> y = np.array([[0, 0], [0, 0], [0, 1], [0, 1], [1, 1], [1, 1], [1, 0], [1, 0]])
    >>> x_train, x_test, y_train, y_test = stratify_train_test_split_multi_label(x, y, test_size=0.5, closest_ratio=False)
    >>> x_train
    array([[1, 2],
           [3, 4],
           [1, 2],
           [3, 4]])
    >>> x_test
    array([[1, 2],
           [3, 4],
           [1, 2],
           [3, 4]])
    >>> y_train
    array([[0, 1],
           [0, 1],
           [1, 0],
           [1, 0]])
    >>> y_test
    array([[1, 1],
           [1, 1],
           [0, 0],
           [0, 0]])
    >>> print("class ratio:", tuple(y_test.sum(0) / y.sum(0)))
    class ratio: (0.5, 0.5)
    >>> print("sample ratio:", y_test.shape[0] / y.shape[0])
    sample ratio: 0.5
    """
    assert len(y.shape) == 2, "y should be 2D"
    assert test_size > 0.0, "test_size cannot be a zero or negative value!"
    x = np.array(x, dtype=np.object_) if not isinstance(x, np.ndarray) else x

    # excluding samples with no objects/tags
    non_objects = np.any(y.sum(1) == 0)
    if non_objects:
        y_no_objects = y[y.sum(1) == 0]
        x_no_objects = x[y.sum(1) == 0]
        x = x[y.sum(1) > 0]
        y = y[y.sum(1) > 0]

    available_samples = np.ones((y.shape[0]), dtype=np.bool8)
    test_samples = np.zeros((y.shape[0]), dtype=np.bool8)
    train_samples = np.zeros((y.shape[0]), dtype=np.bool8)
    class_sample_counts = y.sum(axis=0)
    ideal_train_size = np.floor(sum(class_sample_counts) * (1 - test_size))

    # stratify starts from a class with the lowest number of samples
    class_indices = np.argsort(class_sample_counts)
    for class_index in class_indices:
        test_number_samples = y[:, class_index][test_samples].sum()
        n_test = np.ceil(class_sample_counts[class_index] * test_size)
        n_test = max(0, n_test - test_number_samples)
        input_labels = y[:, class_index].copy()
        input_labels[np.invert(available_samples)] = 0
        if n_test == 0 or len(input_labels) == 0:
            continue
        if closest_ratio:
            chosen_indices, *_ = subset_sum(
                input_numbers=input_labels, target_number=n_test
            )
        else:
            sorted_indices = np.argsort(input_labels)
            cum_sum_values = np.cumsum(input_labels[sorted_indices])
            chosen_indices = sorted_indices[cum_sum_values < n_test].tolist()
            if len(chosen_indices) < len(sorted_indices):
                chosen_indices.append(sorted_indices[len(chosen_indices)])
        # Update available_samples, train_samples, test_samples
        for update_index, n_label in enumerate(input_labels):
            if n_label == 0:
                # samples that have no elements are ignored ...
                continue
            if update_index in chosen_indices:
                test_samples[update_index] = True
                train_samples[update_index] = False
            else:
                test_samples[update_index] = False
                train_samples[update_index] = True
            available_samples[update_index] = False
    # Allocating all the remaining samples to train because the code structure ensures the ratio of test
    # samples to the whole dataset.
    train_samples = np.bitwise_or(train_samples, np.bitwise_not(test_samples))

    if non_objects:
        # splitting samples with no objects trying to save the balance between train and test numbers

        train_left = int(ideal_train_size - sum(train_samples))
        indices = np.arange(len(y_no_objects))
        np.random.shuffle(indices)

        x_no_objects_train, y_no_objects_train = (
            x_no_objects[:train_left],
            y_no_objects[:train_left],
        )
        x_no_objects_test, y_no_objects_test = (
            x_no_objects[train_left:],
            y_no_objects[train_left:],
        )

        return (
            np.concatenate([x[train_samples], x_no_objects_train]),
            np.concatenate([x[test_samples], x_no_objects_test]),
            np.concatenate([y[train_samples], y_no_objects_train]),
            np.concatenate([y[test_samples], y_no_objects_test]),
        )
    else:
        return x[train_samples], x[test_samples], y[train_samples], y[test_samples]



# Google utils_: https://cloud.google.com/storage/docs/reference/libraries

import os
import platform
import subprocess
import time
from pathlib import Path

import torch


def gsutil_getsize(url=""):
    # gs://bucket/file size https://cloud.google.com/storage/docs/gsutil/commands/du
    s = subprocess.check_output("gsutil du %s" % url, shell=True).decode("utf-8")
    return eval(s.split(" ")[0]) if len(s) else 0  # bytes


def attempt_download(weights):
    # Attempt to download pretrained weights if not found locally
    weights = weights.strip().replace("'", "")
    file = Path(weights).name.lower()

    msg = (
        weights
        + " missing, try downloading from https://github.com/ultralytics/yolov5/releases/"
    )
    models = [
        "yolov5s.pt",
        "yolov5m.pt",
        "yolov5l.pt",
        "yolov5x.pt",
    ]  # available models
    redundant = False  # offer second download option

    if file in models and not os.path.isfile(weights):
        # Google Drive
        # d = {'yolov5s.pt': '1R5T6rIyy3lLwgFXNms8whc-387H0tMQO',
        #      'yolov5m.pt': '1vobuEExpWQVpXExsJ2w-Mbf3HJjWkQJr',
        #      'yolov5l.pt': '1hrlqD1Wdei7UT4OgT785BEk1JwnSvNEV',
        #      'yolov5x.pt': '1mM8aZJlWTxOg7BZJvNUMrTnA2AbeCVzS'}
        # r = gdrive_download(id=d[file], name=weights) if file in d else 1
        # if r == 0 and os.path.exists(weights) and os.path.getsize(weights) > 1E6:  # check
        #    return

        try:  # GitHub
            url = "https://github.com/ultralytics/yolov5/releases/download/v3.1/" + file
            print("Downloading %s to %s..." % (url, weights))
            torch.hub.download_url_to_file(url, weights)
            assert os.path.exists(weights) and os.path.getsize(weights) > 1e6  # check
        except Exception as e:  # GCP
            print("Download error: %s" % e)
            assert redundant, "No secondary mirror"
            url = "https://storage.googleapis.com/ultralytics/yolov5/ckpt/" + file
            print("Downloading %s to %s..." % (url, weights))
            r = os.system(
                "curl -L %s -o %s" % (url, weights)
            )  # torch.hub.download_url_to_file(url, weights)
        finally:
            if not (
                os.path.exists(weights) and os.path.getsize(weights) > 1e6
            ):  # check
                os.remove(weights) if os.path.exists(
                    weights
                ) else None  # remove partial downloads
                print("ERROR: Download failure: %s" % msg)
            print("")
            return


def gdrive_download(id="1n_oKgR81BJtqk75b00eAjdv03qVCQn2f", name="coco128.zip"):
    # Downloads a file from Google Drive. from utils_.google_utils import *; gdrive_download()
    t = time.time()

    print(
        "Downloading https://drive.google.com/uc?export=download&id=%s as %s... "
        % (id, name),
        end="",
    )
    os.remove(name) if os.path.exists(name) else None  # remove existing
    os.remove("cookie") if os.path.exists("cookie") else None

    # Attempt file download
    out = "NUL" if platform.system() == "Windows" else "/dev/null"
    os.system(
        'curl -c ./cookie -s -L "drive.google.com/uc?export=download&id=%s" > %s '
        % (id, out)
    )
    if os.path.exists("cookie"):  # large file
        s = (
            'curl -Lb ./cookie "drive.google.com/uc?export=download&confirm=%s&id=%s" -o %s'
            % (get_token(), id, name)
        )
    else:  # small file
        s = 'curl -s -L -o %s "drive.google.com/uc?export=download&id=%s"' % (name, id)
    r = os.system(s)  # execute, capture return
    os.remove("cookie") if os.path.exists("cookie") else None

    # Error check
    if r != 0:
        os.remove(name) if os.path.exists(name) else None  # remove partial
        print("Download error ")  # raise Exception('Download error')
        return r

    # Unzip if archive
    if name.endswith(".zip"):
        print("unzipping... ", end="")
        os.system("unzip -q %s" % name)  # unzip
        os.remove(name)  # remove zip to free space

    print("Done (%.1fs)" % (time.time() - t))
    return r


def get_token(cookie="./cookie"):
    with open(cookie) as f:
        for line in f:
            if "download" in line:
                return line.split()[-1]
    return ""


# def upload_blob(bucket_name, source_file_name, destination_blob_name):
#     # Uploads a file to a bucket
#     # https://cloud.google.com/storage/docs/uploading-objects#storage-upload-object-python
#
#     storage_client = storage.Client()
#     bucket = storage_client.get_bucket(bucket_name)
#     blob = bucket.blob(destination_blob_name)
#
#     blob.upload_from_filename(source_file_name)
#
#     print('File {} uploaded to {}.'.format(
#         source_file_name,
#         destination_blob_name))
#
#
# def download_blob(bucket_name, source_blob_name, destination_file_name):
#     # Uploads a blob from a bucket
#     storage_client = storage.Client()
#     bucket = storage_client.get_bucket(bucket_name)
#     blob = bucket.blob(source_blob_name)
#
#     blob.download_to_filename(destination_file_name)
#
#     print('Blob {} downloaded to {}.'.format(
#         source_blob_name,
#         destination_file_name))



from logging import Logger
from typing import Union, Dict
from deep_utils.utils.logging_utils import log_print, value_error_log


class MinIOUtils:
    @staticmethod
    def minio_get(minio_client, bucket_name, object_name, logger: Union[None, Logger]):
        """
        Get object from minio_client. This function is created for compatibility otherwise no extra functionality provided compared to the main module.
        :param minio_client:
        :param bucket_name:
        :param object_name:
        :param logger:
        :return:
        """
        obj = minio_client.get_object(bucket_name, object_name)
        log_print(
            logger=logger,
            message=f"Successfully got object: {object_name} from bucket: {bucket_name}",
        )
        return obj

    @staticmethod
    def minio_fget(
            minio_client, bucket_name, object_name, file_path, logger: Union[None, Logger]
    ):
        """
        Get file object from minio_client. This function is created for compatibility otherwise no extra functionality provided compared to the main module.
        :param minio_client:
        :param bucket_name:
        :param object_name:
        :param file_path: Where to save the file
        :param logger:
        :return:
        """
        obj = minio_client.fget_object(bucket_name, object_name, file_path)
        log_print(
            logger=logger,
            message=f"Successfully got object: {object_name} from bucket: {bucket_name}",
        )
        return obj

    @staticmethod
    def minio_put(
            minio_client,
            bucket_name,
            object_name,
            data,
            create=True,
            logger: Union[Logger, None] = None,
    ):
        """
        put an object inside a bucket
        :param minio_client:
        :param bucket_name:
        :param object_name:
        :param data: data-file to be stored in
        :param create: create te bucket in case it doesn't exist, default is True.
        :param logger: a logger instance
        :return:
        """
        MinIOUtils.create_bucket(minio_client, bucket_name, create, logger, )
        try:
            length = len(data.read())
            data.seek(0)
            part_size = 0
        except:
            length = -1
            part_size = data.__sizeof__()
        result = minio_client.put_object(
            bucket_name,
            object_name,
            data,
            length=length,
            part_size=part_size
        )
        log_print(
            logger=logger,
            message=f"Successfully put object: {object_name} into bucket: {bucket_name}",
        )
        return result

    @staticmethod
    def minio_fput(
            minio_client,
            bucket_name,
            object_name,
            file_path,
            create=True,
            logger: Union[Logger, None] = None,
    ):
        """
        put a file inside a bucket
        :param minio_client:
        :param bucket_name:
        :param object_name:
        :param file_path: file to be stored in the given bucket
        :param create: create te bucket in case it doesn't exist, default is True.
        :param logger: a logger instance
        :return:
        """
        MinIOUtils.create_bucket(minio_client, bucket_name, create, logger)

        result = minio_client.fput_object(
            bucket_name,
            object_name,
            file_path,
        )
        log_print(
            logger=logger,
            message=f"Successfully put file: {object_name} into bucket: {bucket_name}",
        )
        return result

    @staticmethod
    def create_bucket(minio_client, bucket_name, create=True, logger=None) -> bool:
        """
        This method is used to find the requested bucket and create it in case the user desires it.
        :param bucket_name:
        :param create:
        :param logger:
        :param minio_client:
        :return:
        """
        found = minio_client.bucket_exists(bucket_name)
        if found:
            log_print(logger=logger,
                      message=f"Bucket {bucket_name} already exists")
        elif not found and create:
            minio_client.make_bucket(bucket_name)
            log_print(logger=logger,
                      message=f"Successfully Created bucket: {bucket_name}")
        else:
            value_error_log(
                logger=logger, message='f"Error in creating minio bucket"')
        return found

    @staticmethod
    def check_minio_connection(minio_host, status_key="MINIO_STATUS") -> Dict[str, str]:
        import requests
        # check minio
        status = dict()
        try:
            minio_status = requests.get(f"http://{minio_host}/minio/health/live").ok
            status[status_key] = "Alive" if minio_status else "Down"
        except:
            status[status_key] = "Down"
        return status

    @staticmethod
    def _get_minio_policy(bucket_name):
        policy = {"Statement": [{"Action": ["s3:GetBucketLocation"],
                                 "Effect": "Allow", "Principal": {"AWS": ["*"]},
                                 "Resource": [f"arn:aws:s3:::{bucket_name}"]},
                                {"Action": ["s3:GetObject"],
                                 "Effect": "Allow", "Principal": {"AWS": ["*"]},
                                 "Resource": [f"arn:aws:s3:::{bucket_name}/*"]}], "Version": "2012-10-17"}
        return policy

    @staticmethod
    def make_bucket_public(minio_client, bucket_name, logger=None, verbose=1):
        import json
        found = MinIOUtils.create_bucket(minio_client, bucket_name)
        if not found:
            minio_client.set_bucket_policy(bucket_name, json.dumps(MinIOUtils._get_minio_policy(bucket_name)))
            log_print(logger=logger, message=f"Successfully Made bucket: {bucket_name} public", verbose=verbose)



from .torch.deep_sort import DeepSortTorch, DeepSortTorchFeatureExtractor, DeepSortTorchTracker

__all__ = ["DeepSortTorch", "build_tracker", "DeepSortTorchFeatureExtractor", "DeepSortTorchTracker"]


def build_tracker(cfg, use_cuda):
    return DeepSortTorch(
        cfg.DEEPSORT.REID_CKPT,
        max_dist=cfg.DEEPSORT.MAX_DIST,
        min_confidence=cfg.DEEPSORT.MIN_CONFIDENCE,
        nms_max_overlap=cfg.DEEPSORT.NMS_MAX_OVERLAP,
        max_iou_distance=cfg.DEEPSORT.MAX_IOU_DISTANCE,
        max_age=cfg.DEEPSORT.MAX_AGE,
        n_init=cfg.DEEPSORT.N_INIT,
        nn_budget=cfg.DEEPSORT.NN_BUDGET,
        use_cuda=use_cuda,
    )



class frozendict:
    def __init__(self, **kwargs):
        super(frozendict, self).__init__()
        self.repo = dict()
        for key, item in kwargs.items():
            self.repo[key] = item

    def __getitem__(self, item):
        return self.repo[item]

    def __setitem__(self, key, value):
        raise TypeError(f"frozendict object does not support updating")


def get_dict_extreme(dict_: dict, key, mode='max', list_values=False, ignore_non_sequences=True):
    """
    Gets the maximum/minimum item of a dict based on a key in the input dict
    :param dict_:
    :param key:
    :param mode:
    :param list_values: whether list values, so it would have the same shape as the input
    :param ignore_non_sequences: In the input dict, there may be some features which are not a sequence type and do no
    accept index. Ignore if those samples are seen.
    :return:
    """
    import numpy as np
    if mode == 'max':
        index = np.argmax(dict_[key])
    elif mode == 'min':
        index = np.argmax(dict_[key])
    else:
        raise ValueError()
    res = dict()

    for k, v in dict_.items():
        try:
            value = v[index]
        except TypeError as t:
            if ignore_non_sequences:
                value = v
            else:
                raise TypeError(t)
        if list_values:
            res[k] = [value]
        else:
            res[k] = value

    return res



try:
    from deep_utils.dummy_objects.utils.memory_utils import MemoryUtilsTorch
    from deep_utils.utils.memory_utils.torch_memory_utils import MemoryUtilsTorch
except ModuleNotFoundError:
    pass



import numpy as np
import tensorflow as tf


def post_processing(
    reg_list,
    cls_list,
    num_classes,
    image_size,
    feature_map_wh_list,
    min_boxes,
    center_variance,
    size_variance,
    conf_threshold=0.6,
    nms_max_output_size=100,
    nms_iou_threshold=0.3,
    top_k=100,
):
    reg_list = [tf.keras.layers.Reshape([-1, 4])(reg) for reg in reg_list]
    cls_list = [tf.keras.layers.Reshape(
        [-1, num_classes])(cls) for cls in cls_list]

    reg = tf.keras.layers.Concatenate(axis=1)(reg_list)
    cls = tf.keras.layers.Concatenate(axis=1)(cls_list)

    # post process
    cls = tf.keras.layers.Softmax(axis=-1)(cls)
    loc = decode_regression(
        reg, image_size, feature_map_wh_list, min_boxes, center_variance, size_variance
    )

    result = tf.keras.layers.Concatenate(axis=-1)([cls, loc])

    # confidence thresholding
    mask = conf_threshold < cls[..., 1]
    result = tf.boolean_mask(tensor=result, mask=mask)

    # non-maximum suppression
    mask = tf.image.non_max_suppression(
        boxes=result[..., -4:],
        scores=result[..., 1],
        max_output_size=nms_max_output_size,
        iou_threshold=nms_iou_threshold,
        name="non_maximum_suppresion",
    )
    result = tf.gather(params=result, indices=mask, axis=0)

    # top-k filtering
    top_k_value = tf.math.minimum(tf.constant(top_k), tf.shape(result)[0])
    mask = tf.nn.top_k(result[..., 1], k=top_k_value, sorted=True).indices
    result = tf.gather(params=result, indices=mask, axis=0)

    return result


def decode_regression(
    reg, image_size, feature_map_w_h_list, min_boxes, center_variance, size_variance
):
    priors = []
    for feature_map_w_h, min_box in zip(feature_map_w_h_list, min_boxes):
        xy_grid = np.meshgrid(
            range(feature_map_w_h[0]), range(feature_map_w_h[1]))
        xy_grid = np.add(xy_grid, 0.5)
        xy_grid[0, :, :] /= feature_map_w_h[0]
        xy_grid[1, :, :] /= feature_map_w_h[1]
        xy_grid = np.stack(xy_grid, axis=-1)
        xy_grid = np.tile(xy_grid, [1, 1, len(min_box)])
        xy_grid = np.reshape(xy_grid, (-1, 2))

        wh_grid = np.array(min_box) / np.array(image_size)[:, np.newaxis]
        wh_grid = np.tile(np.transpose(wh_grid), [
                          np.product(feature_map_w_h), 1])

        prior = np.concatenate((xy_grid, wh_grid), axis=-1)
        priors.append(prior)

    priors = np.concatenate(priors, axis=0)

    priors = tf.constant(priors, dtype=tf.float32,
                         shape=priors.shape, name="priors")

    center_xy = reg[..., :2] * center_variance * \
        priors[..., 2:] + priors[..., :2]
    center_wh = tf.exp(reg[..., 2:] * size_variance) * priors[..., 2:]

    # center to corner
    start_xy = center_xy - center_wh / 2
    end_xy = center_xy + center_wh / 2

    loc = tf.concat([start_xy, end_xy], axis=-1)
    loc = tf.clip_by_value(loc, clip_value_min=0.0, clip_value_max=1.0)

    return loc



import os
from os.path import join


def unzip(zip_path, zip_dir="."):
    from zipfile import ZipFile

    with ZipFile(zip_path, "r") as zip:
        print(f"extracting {zip_path}")
        zip.extractall(zip_dir)
        print(f"extracting is done!")


def unzip_dir_zip(dir_path, remove_zip_file=False):
    """
    Unzip a directory of zip files
    :param dir_path: path to directory that contains zip files
    :param remove_zip_file: whether to remove zip-file, default is False.
    :return:
    """
    for base, directories, file_names in os.walk(dir_path):
        for file_name in file_names:
            if file_name.endswith(".zip"):
                zip_path = join(base, file_name)
                unzip(zip_path, zip_dir=base)
                if remove_zip_file:
                    os.system(f'rm -rf "{zip_path}"')



from collections import OrderedDict

from deep_utils.utils.list_utils import shift_lst


class BlocksTF:
    @staticmethod
    def conv_norm_act(
            out_c,
            k=(3, 3),
            s=(1, 1),
            p="valid",
            norm="bn",
            act="relu",
            conv=True,
            index=0,
            conv_kwargs: dict = None,
            norm_kwargs: dict = None,
            act_kwargs: dict = None,
            pooling=None,
            pooling_k=(2, 2),
            pooling_s=(2, 2),
            pool_kwargs=None,
            move_forward=0,
    ):
        from tensorflow import keras

        conv_kwargs = dict() if conv_kwargs is None else conv_kwargs
        norm_kwargs = dict() if norm_kwargs is None else norm_kwargs
        act_kwargs = dict() if act_kwargs is None else act_kwargs
        pool_kwargs = dict() if pool_kwargs is None else pool_kwargs
        k = (k, k) if isinstance(k, int) else k
        s = (s, s) if isinstance(s, int) else s
        p = (p, p) if isinstance(p, int) else p
        pooling_s = (pooling_s, pooling_s) if isinstance(
            pooling_s, int) else pooling_s
        pooling_k = (pooling_k, pooling_k) if isinstance(
            pooling_k, int) else pooling_k
        modules, names = [], []
        if conv:
            modules.append(BlocksTF.conv_2d(out_c, k, s, p, **conv_kwargs))
            names.append(f"conv2d_{index}_f{out_c}_k{k[0]}_s{s[0]}")
        if norm:
            modules.append(BlocksTF.load_layer_norm(norm, **norm_kwargs))
            names.append(f"{norm}2d_{index}")
        if act:
            modules.append(BlocksTF.load_activation(act, **act_kwargs))
            names.append(f"{act}_{index}")
        modules = shift_lst(modules, move_forward=move_forward % 3)
        names = shift_lst(names, move_forward=move_forward % 3)
        if pooling:
            pooling_layer = BlocksTF.load_pooling(
                pooling, pooling_k, pooling_s, **pool_kwargs
            )
            modules.append(pooling_layer)
            names.append(f"{pooling}2d_{index}")
        layer_dict = OrderedDict(
            {name: module for name, module in zip(names, modules)})
        cnn = keras.Sequential()
        for value in layer_dict.values():
            cnn.add(value)
        return cnn

    @staticmethod
    def load_pooling(pooling, k, s, **pooling_kwargs):
        from tensorflow.keras import layers

        if pooling == "max":
            pooling = layers.MaxPooling2D(
                pool_size=k, strides=s, **pooling_kwargs)
        elif pooling == "avg":
            pooling = layers.AveragePooling2D(
                pool_size=k, strides=s, **pooling_kwargs)
        else:
            raise ValueError(
                f"requested pooling {pooling} is not supported. Supported poolings: max, avg"
            )
        return pooling

    @staticmethod
    def load_activation(act, **act_kwargs):
        from tensorflow.keras import layers

        if act == "relu":
            activation = layers.ReLU(**act_kwargs)
        elif act == "leaky_relu":
            activation = layers.LeakyReLU(**act_kwargs)
        else:
            raise ValueError(
                f"requested activation {act} is not supported. Supported activations: relu, leaky_relu"
            )
        return activation

    @staticmethod
    def conv_2d(out_c, k=(3, 3), s=(1, 1), padding="valid", bias=True):
        from tensorflow.keras import layers

        return layers.Conv2D(
            filters=out_c, kernel_size=k, strides=s, padding=padding, use_bias=bias
        )

    @staticmethod
    def load_layer_norm(norm, **norm_kwargs):
        from tensorflow.keras import layers

        if norm == "bn":
            layer_norm = layers.BatchNormalization(**norm_kwargs)
        elif norm == "ln":
            layer_norm = layers.LayerNormalization(**norm_kwargs)
        else:
            raise ValueError(
                f"requested layer_norm {norm} is not supported. Supported layer_norms: bn(batch normalization), "
                f"gn(group normalization"
            )
        return layer_norm

    @staticmethod
    def res_basic_block(in_c, out_c, bias, down_sample=False):
        from tensorflow import keras

        modules = []
        names = []
        for i in range(2):
            modules.append(
                BlocksTF.conv_norm_act(
                    int(in_c / 2) if down_sample and i == 0 else in_c,
                    out_c,
                    act=False if i == 1 else "relu",
                    s=(2, 2) if down_sample and i == 0 else (1, 1),
                    conv_kwargs={"bias": bias},
                )
            )
            names.append(f"cnn_{i}_in{in_c}_f{out_c}")
        res_block = keras.Sequential(
            OrderedDict({name: module for name, module in zip(names, modules)})
        )
        return res_block


if __name__ == "__main__":
    a = BlocksTF.conv_norm_act(32, 64, move_forward=2)
    a.build((1, None, None, 3))
    print(a.summary())



import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision.models._utils as _utils

from .net import FPN as FPN
from .net import SSH as SSH
from .net import MobileNetV1 as MobileNetV1


class ClassHead(nn.Module):
    def __init__(self, inchannels=512, num_anchors=3):
        super(ClassHead, self).__init__()
        self.num_anchors = num_anchors
        self.conv1x1 = nn.Conv2d(
            inchannels, self.num_anchors * 2, kernel_size=(1, 1), stride=1, padding=0
        )

    def forward(self, x):
        out = self.conv1x1(x)
        out = out.permute(0, 2, 3, 1).contiguous()

        return out.view(out.shape[0], -1, 2)


class BboxHead(nn.Module):
    def __init__(self, inchannels=512, num_anchors=3):
        super(BboxHead, self).__init__()
        self.conv1x1 = nn.Conv2d(
            inchannels, num_anchors * 4, kernel_size=(1, 1), stride=1, padding=0
        )

    def forward(self, x):
        out = self.conv1x1(x)
        out = out.permute(0, 2, 3, 1).contiguous()

        return out.view(out.shape[0], -1, 4)


class LandmarkHead(nn.Module):
    def __init__(self, inchannels=512, num_anchors=3):
        super(LandmarkHead, self).__init__()
        self.conv1x1 = nn.Conv2d(
            inchannels, num_anchors * 10, kernel_size=(1, 1), stride=1, padding=0
        )

    def forward(self, x):
        out = self.conv1x1(x)
        out = out.permute(0, 2, 3, 1).contiguous()

        return out.view(out.shape[0], -1, 10)


class RetinaFace(nn.Module):
    def __init__(self, cfg=None, phase="test", pretrained_model=None):
        """
        :param cfg:  Network related settings.
        :param phase: train or test.
        """
        super(RetinaFace, self).__init__()
        self.phase = phase
        backbone = None
        if cfg["name"] == "mobilenet":
            backbone = MobileNetV1()
            if cfg["pretrain"]:
                checkpoint = torch.load(
                    pretrained_model, map_location=torch.device("cpu")
                )
                from collections import OrderedDict

                new_state_dict = OrderedDict()
                for k, v in checkpoint["state_dict"].items():
                    name = k[7:]  # remove module.
                    new_state_dict[name] = v
                # load params
                backbone.load_state_dict(new_state_dict)
        elif cfg["name"] == "resnet50":
            import torchvision.models as models

            backbone = models.resnet50(pretrained=cfg["pretrain"])

        self.body = _utils.IntermediateLayerGetter(
            backbone, cfg["return_layers"])
        in_channels_stage2 = cfg["in_channel"]
        in_channels_list = [
            in_channels_stage2 * 2,
            in_channels_stage2 * 4,
            in_channels_stage2 * 8,
        ]
        out_channels = cfg["out_channel"]
        self.fpn = FPN(in_channels_list, out_channels)
        self.ssh1 = SSH(out_channels, out_channels)
        self.ssh2 = SSH(out_channels, out_channels)
        self.ssh3 = SSH(out_channels, out_channels)

        self.ClassHead = self._make_class_head(
            fpn_num=3, inchannels=cfg["out_channel"])
        self.BboxHead = self._make_bbox_head(
            fpn_num=3, inchannels=cfg["out_channel"])
        self.LandmarkHead = self._make_landmark_head(
            fpn_num=3, inchannels=cfg["out_channel"]
        )

    def _make_class_head(self, fpn_num=3, inchannels=64, anchor_num=2):
        classhead = nn.ModuleList()
        for i in range(fpn_num):
            classhead.append(ClassHead(inchannels, anchor_num))
        return classhead

    def _make_bbox_head(self, fpn_num=3, inchannels=64, anchor_num=2):
        bboxhead = nn.ModuleList()
        for i in range(fpn_num):
            bboxhead.append(BboxHead(inchannels, anchor_num))
        return bboxhead

    def _make_landmark_head(self, fpn_num=3, inchannels=64, anchor_num=2):
        landmarkhead = nn.ModuleList()
        for i in range(fpn_num):
            landmarkhead.append(LandmarkHead(inchannels, anchor_num))
        return landmarkhead

    def forward(self, inputs):
        out = self.body(inputs)

        # FPN
        fpn = self.fpn(out)

        # SSH
        feature1 = self.ssh1(fpn[0])
        feature2 = self.ssh2(fpn[1])
        feature3 = self.ssh3(fpn[2])
        features = [feature1, feature2, feature3]

        bbox_regressions = torch.cat(
            [self.BboxHead[i](feature) for i, feature in enumerate(features)], dim=1
        )
        classifications = torch.cat(
            [self.ClassHead[i](feature) for i, feature in enumerate(features)], dim=1
        )
        ldm_regressions = torch.cat(
            [self.LandmarkHead[i](feature) for i, feature in enumerate(features)], dim=1
        )

        if self.phase == "train":
            output = (bbox_regressions, classifications, ldm_regressions)
        else:
            output = (
                bbox_regressions,
                F.softmax(classifications, dim=-1),
                ldm_regressions,
            )
        return output



from deep_utils.utils.logging_utils import log_print


class TensorboardTorch:
    def __init__(
        self,
        log_dir,
        comment="",
        purge_step=None,
        max_queue=10,
        flush_secs=120,
        filename_suffix="",
        logger=None,
        verbose=1,
    ):
        from torch.utils.tensorboard import SummaryWriter

        self.writer = SummaryWriter(
            log_dir=log_dir,
            comment=comment,
            purge_step=purge_step,
            max_queue=max_queue,
            flush_secs=flush_secs,
            filename_suffix=filename_suffix,
        )
        log_print(
            logger,
            f"Successfully created TensorboardTorch that saves to {log_dir}",
            verbose=verbose,
        )

    def __call__(self, epoch: int, **logs):
        for metric_name, metric_value in logs.items():
            self.writer.add_scalar(metric_name, metric_value, epoch)



from collections import defaultdict
from pathlib import Path
from typing import Union

import numpy as np

from deep_utils.utils.logging_utils import log_print


class CSVLogger:
    def __init__(self, csv_path: Union[Path, str], logger=None, verbose=1):
        self.csv_path = csv_path
        self.logs: dict = defaultdict(list)
        log_print(
            logger,
            f"Successfully created CSVLogger that saves to {self.csv_path}",
            verbose=verbose,
        )

    def __call__(self, epoch=None, **logs):
        for key, val in logs.items():
            self.logs[key].append(val)
        self.__save()

    def __save(self):
        import pandas as pd

        columns = list(self.logs.keys())
        values = np.array(list(self.logs.values())).T
        df = pd.DataFrame(values, columns=columns)
        df.to_csv(self.csv_path)



import numpy as np


def read_img_form(contents, ret_rgb: bool = False):
    """
    read input from a form request with cv2.
    :param contents: The parsed form input
    :param ret_rgb: If set true, will change the output to rgb
    :return: cv2 - img
    """
    import cv2

    image = np.array(bytearray(contents.read()), dtype=np.uint8)
    image = cv2.imdecode(image, cv2.IMREAD_COLOR)
    if ret_rgb:
        # convert to rgb
        image = image[..., ::-1]
    return image



import warnings

import pytest

warnings.filterwarnings("ignore")


@pytest.mark.basic
def test_resize_ratio():
    """
    Testing the resize_ratio function
    Returns: Nothing

    """
    import numpy as np

    from deep_utils import resize_ratio

    dummy_images = [
        np.random.randint(0, 255, (1200, 900, 3), dtype=np.uint8),
        np.random.randint(0, 255, (800, 1200, 3), dtype=np.uint8),
        np.random.randint(0, 255, (550, 350, 3), dtype=np.uint8),
        np.random.randint(0, 255, (900, 900, 3), dtype=np.uint8),
    ]
    preferred_outputs = [
        (900, 675, 3),
        (600, 900, 3),
        (900, 572, 3),
        (900, 900, 3),
    ]

    for dummy_img, preferred_output in zip(dummy_images, preferred_outputs):
        out = resize_ratio(dummy_img, 900)
        assert (
            out.shape == preferred_output
        ), f"resize_ratio failed for input img.shape={dummy_img.shape}"



import numpy as np

from deep_utils.utils.box_utils.boxes import Box
from deep_utils.utils.logging_utils import log_print
from deep_utils.utils.resize_utils.main_resize import resize, resize_ratio


def get_horizontal_image_stack(images, size):
    widths, heights = [], []
    for img in images:
        w, h = img.shape[:2]
        widths.append(w)
        heights.append(h)
    width = sum(widths)
    height = max(heights)
    img_base = np.zeros((width, height, 3), dtype=np.uint8)
    w_ = 0
    for img in images:
        w, h = img.shape[:2]
        img_base[w_:w_ + w, 0: h] = img
        w_ += w
    if width > size or height > size:
        img_base = resize_ratio(img_base, size)
    return img_base


def get_grid_images(
        images,
        size=(128, 128),
        n_channels=3,
        texts=None,
        text_org=None,
        text_kwargs=None,
        title=None,
        title_org=None,
        title_kwargs=None,
):
    """
    Create a group of images in a grid! Returns a numpy array.
    Args:
        images:
        size:
        n_channels:
        texts:
        text_org:
        text_kwargs:
        title:
        title_org:
        title_kwargs:

    Returns:

    """
    import math

    title_kwargs = dict() if title_kwargs is None else title_kwargs
    text_kwargs = dict() if text_kwargs is None else text_kwargs
    n_images = len(images)
    n = math.ceil(math.sqrt(len(images)))
    columns = n
    rows = math.ceil(n_images / columns)
    img_size = (rows * size[0], columns * size[1])
    img = np.zeros((img_size[0], img_size[1], n_channels), dtype=np.uint8)
    i = 0
    for r in range(rows):
        for c in range(columns):
            resized_img = resize(images[i], size)
            if texts is not None:
                org = (
                    (size[1] // 10, size[0] // 2 - size[0] // 5)
                    if text_org is None
                    else text_org[i]
                )
                resized_img = Box.put_text(
                    resized_img, texts[i], org, **text_kwargs)
            img[
            r * size[0]: (r + 1) * size[0], c * size[1]: (c + 1) * size[1]
            ] = resized_img
            i += 1
            if i == len(images):
                break
        if i == len(images):
            break
    img = np.concatenate([np.zeros((size[0] // 5, img_size[1], 3)), img])
    if title:
        if title_org is None:
            title_org = (
                (size[1] // 10, img_size[0] // 2 - img_size[0] // 5)
                if text_org is None
                else text_org[i]
            )
        img = Box.put_text(img, text=title, org=title_org, **title_kwargs)
    img = img.astype(np.uint8)
    return img


def get_segmentation_grid_image(data_loader, save_path, n_samples=10, logger=None):
    """
    Save and get Segmentation Batches! One row is responsible for train images and the other one to mask images.
    """
    import os

    import cv2

    os.makedirs(save_path, exist_ok=True)
    c = 0
    for en, (x, y) in enumerate(data_loader):
        images = []
        for x_, y_ in zip(x, y):
            images.append(x_ * 255)
            images.append(np.stack((y_[..., 0] * 255,) * 3, axis=-1))
        img = get_grid_images(images)
        cv2.imwrite(os.path.join(
            save_path, f"batch_samples_{c}.jpg"), img[..., ::-1])
        c += 1
        if c >= n_samples:
            break
    log_print(logger, "Successfully visualized input samples!")



import cv2
import pytest

from deep_utils import Box, YOLOV5TorchObjectDetector, download_file


@pytest.mark.torchvision
def test_detection():
    model = YOLOV5TorchObjectDetector(model_weight="yolov5s.pt", device="cpu")
    file_path = download_file(
        "https://raw.githubusercontent.com/pooya-mohammadi/deep-utils-notebooks/main/vision/images/dog.jpg"
    )
    img = cv2.imread(file_path)
    res = model.detect_objects(img, is_rgb=False, get_time=True)
    assert len(res.boxes) >= 0, "Model did not detect anything!"
    img = Box.put_box_text(img, res.boxes, res.class_names)
    assert img is not None, "put_box_text returned nothing!"



import inspect
import time
from functools import wraps
from typing import Sequence, Iterable

import numpy as np


def get_from_config(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        config = self.config
        arguments = inspect.getfullargspec(func)
        if arguments.defaults is not None:
            kwargs_ = {
                i: j for i, j in zip(arguments.args[::-1], arguments.defaults[::-1])
            }
        else:
            kwargs_ = dict()
        if kwargs is not None:
            kwargs_.update(kwargs)
        kwargs = kwargs_
        for key, val in kwargs.items():
            if val is None and hasattr(config, key):
                val = getattr(config, key)
                kwargs[key] = val
        return func(self, *args, **kwargs)

    return wrapper


def expand_input(dim):
    def inner_decorator(func):
        @wraps(func)
        def wrapper(self, in_, *args, **kwargs):
            if isinstance(in_, list) and np.all([len(input_obj.shape) == dim for input_obj in in_]):
                # if there is a list of inputs and each input is a ndarray with the shape of dim...
                return func(self, in_, *args, **kwargs)
            if len(in_.shape) == dim + 1:
                return func(self, in_, *args, **kwargs)
            elif len(in_.shape) == dim:
                in_ = np.expand_dims(in_, axis=0)
                results = func(self, in_, *args, **kwargs)
                if hasattr(results, "DictNamedTuple") and results.DictNamedTuple:
                    new_results = {
                        key: val[0] if val is not None and isinstance(val, Iterable) and len(val) == 1 else val
                        for key, val in results.items()
                    }
                    cls = type(results)
                    results = cls(**new_results)
                elif isinstance(results, tuple):
                    results = tuple(
                        [
                            res[0] if res is not None and len(
                                res) == 1 else res
                            for res in results
                        ]
                    )
                elif isinstance(results, dict):
                    results = {
                        key: val[0] if val is not None and len(
                            val) == 1 else val
                        for key, val in results.items()
                    }
                else:
                    results = results[0]
                return results
            else:
                raise Exception(f"shape {in_.shape} is not supported.")

        return wrapper

    return inner_decorator


def get_elapsed_time(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        elapsed_time = kwargs.get("get_time", False)
        if elapsed_time:
            tic = time.time()
            results = func(self, *args, **kwargs)
            toc = time.time()
            elapsed_time = round(toc - tic, 4)
            if isinstance(results, dict):
                results["elapsed_time"] = elapsed_time
            elif hasattr(results, "DictNamedTuple") and results.DictNamedTuple:
                from deep_utils.utils.dict_named_tuple_utils import dictnamedtuple

                new_results = dict(results.items())
                new_results["elapsed_time"] = elapsed_time
                cls = dictnamedtuple(
                    results.TypeName, list(results._fields) + ["elapsed_time"]
                )
                results = cls(**new_results)
            else:
                results = tuple(list(results) + [elapsed_time])
            return results
        return func(self, *args, **kwargs)

    return wrapper


def rgb2bgr(in_):
    def inner_decorator(func):
        @wraps(func)
        def wrapper(self, in_img, *args, **kwargs):
            is_rgb = kwargs.get("is_rgb")
            # if not is_rgb and in_ == 'rgb':
            #     in_img = in_img[..., ::-1]
            # elif is_rgb and in_ == 'bgr':
            #     in_img = in_img[..., ::-1]
            # elif is_rgb and in_ == 'gray':
            #     in_img = np.dot(in_img[..., :3], [0.299, 0.587, 0.144])
            #     in_img = in_img.astype(np.uint8)
            # elif not is_rgb and in_ == 'gray':
            #     in_img = np.dot(in_img[..., :3][..., ::-1], [0.299, 0.587, 0.144])
            #     in_img = in_img.astype(np.uint8)
            in_img = lib_rgb2bgr(in_img, target_type=in_, is_rgb=is_rgb)
            return func(self, in_img, *args, **kwargs)

        return wrapper

    return inner_decorator


def lib_rgb2bgr(in_img, target_type, is_rgb):
    if not is_rgb and target_type == "rgb":
        in_img = in_img[..., ::-1]
    elif (is_rgb and target_type == "rgb") or (not is_rgb and target_type == 'bgr'):
        pass
    elif is_rgb and target_type == "bgr":
        in_img = in_img[..., ::-1]
    elif is_rgb and target_type == "gray":
        in_img = np.dot(in_img[..., :3], [0.299, 0.587, 0.144])
        in_img = in_img.astype(np.uint8)
    elif not is_rgb and target_type == "gray":
        in_img = np.dot(in_img[..., :3][..., ::-1], [0.299, 0.587, 0.144])
        in_img = in_img.astype(np.uint8)
    else:
        raise ValueError(
            f"[ERROR] The input target_type: {target_type} and is_rgb: {is_rgb} are not supported!"
        )
    return in_img


def in_shape_fix(in_: np.ndarray, size=4):
    if len(in_.shape) == size:
        return in_
    elif len(in_.shape) == size - 1:
        return np.expand_dims(in_, axis=0)
    else:
        raise ValueError(f"[Error] Input size: {in_.shape} is not valid!")


def out_shape_fix(results):
    if hasattr(results, "DictNamedTuple") and results.DictNamedTuple:
        new_results = {
            key: val[0]
            if val is not None and isinstance(val, Sequence) and len(val) == 1
            else val
            for key, val in results.items()
        }
        cls = type(results)
        results = cls(**new_results)
    elif isinstance(results, tuple):
        results = tuple(
            [
                val[0]
                if val is not None and isinstance(val, Sequence) and len(val) == 1
                else val
                for val in results
            ]
        )
    elif isinstance(results, dict):
        results = {
            key: val[0]
            if val is not None and isinstance(val, Sequence) and len(val) == 1
            else val
            for key, val in results.items()
        }
    else:
        results = results[0]
    return results


def cast_kwargs_dict(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        arguments = inspect.getfullargspec(func)
        for kwarg in arguments.args:
            if kwarg.endswith("_kwargs"):
                _kwargs = kwargs.get(kwarg, None)
                kwargs[kwarg] = dict() if _kwargs is None else _kwargs
        return func(self, *args, **kwargs)

    return wrapper



from typing import List, Tuple, Union


def random_choice_group(
    group_1: Union[List, Tuple], group_2: Union[List, Tuple]
) -> tuple:
    assert len(group_1) == len(
        group_2), "lengths of the input groups are not the same!"
    from random import choice

    index = choice(list(range(len(group_1))))
    return group_1[index], group_2[index]



from sqlalchemy import select, and_, or_
from sqlalchemy.exc import NoResultFound
from typing import Dict, Any


class SQLAlchemyChecks:

    @staticmethod
    def check(cls, key_value: Dict[str, Any], combine="and", session=None) -> bool:

        conditions = [getattr(cls, key) == val for key, val in key_value.items()]
        condition = and_(*conditions) if combine.lower() == "and" else or_(*conditions)

        selected = select(cls).where(condition)
        try:
            session.scalars(selected).one()
        except NoResultFound:
            return True
        return False



import sys
import numpy as np
import torch
from torchreid.utils import FeatureExtractor
from deep_utils.utils.color_utils.color_utils import Colors
from deep_utils.utils.box_utils.boxes import Box
from .sort.detection import Detection
from .sort.nn_matching import NearestNeighborDistanceMetric
from .sort.tracker import Tracker

sys.path.append("deep_sort/deep/reid")


class DeepSortTorchTracker:
    def __init__(self,
                 max_dist=0.2,
                 max_iou_distance=0.7,
                 max_age=70,
                 n_init=3,
                 nn_budget=100):
        max_cosine_distance = max_dist
        metric = NearestNeighborDistanceMetric(
            "euclidean", max_cosine_distance, nn_budget
        )
        self.tracker = Tracker(
            metric, max_iou_distance=max_iou_distance, max_age=max_age, n_init=n_init
        )
        self.colors = Colors()

    def visualize_output_image(self, img, update_output, confidences, names):
        if len(update_output) > 0:
            for j, (output, conf) in enumerate(zip(update_output, confidences)):
                bboxes = output[0:4]
                id = output[4]
                cls = output[5]

                c = int(cls)  # integer class
                label = f'{id} {names[c]} {conf:.2f}'
                boxes = np.array(Box.box2box(bboxes, to_source=Box.BoxSource.Numpy, in_source=Box.BoxSource.CV))
                color = self.colors(c, True)
                img = Box.put_box_text(img, boxes, label, color=color)
        return img

    def update(self, features, classes, confidences, bbox_xywh, height, width, use_yolo_preds=False):
        # self.height, self.width = height, width
        bbox_tlwh = self._xywh_to_tlwh(bbox_xywh)
        detections = [
            Detection(bbox_tlwh[i], conf, features[i])
            for i, conf in enumerate(confidences)
        ]
        # update tracker
        self.tracker.predict()
        self.tracker.update(detections, classes)

        # output bbox identities
        outputs = []
        for track in self.tracker.tracks:
            if not track.is_confirmed() or track.time_since_update > 1:
                continue
            if use_yolo_preds:
                det = track.get_yolo_pred()
                x1, y1, x2, y2 = self._tlwh_to_xyxy(det.tlwh, height, width)
            else:
                box = track.to_tlwh()
                x1, y1, x2, y2 = self._tlwh_to_xyxy(box, height, width)
            track_id = track.track_id
            class_id = track.class_id
            outputs.append(
                np.array([x1, y1, x2, y2, track_id, class_id], dtype=np.int))
        if len(outputs) > 0:
            outputs = np.stack(outputs, axis=0)
        return outputs

    @staticmethod
    def _tlwh_to_xyxy(bbox_tlwh, height, width):
        """
        TODO:
            Convert bbox from xtl_ytl_w_h to xc_yc_w_h
        Thanks JieChen91@github.com for reporting this bug!
        """
        x, y, w, h = bbox_tlwh
        x1 = max(int(x), 0)
        x2 = min(int(x + w), width - 1)
        y1 = max(int(y), 0)
        y2 = min(int(y + h), height - 1)
        return x1, y1, x2, y2

    @staticmethod
    def _xywh_to_tlwh(bbox_xywh):
        if isinstance(bbox_xywh, np.ndarray):
            bbox_tlwh = bbox_xywh.copy()
        elif isinstance(bbox_xywh, torch.Tensor):
            bbox_tlwh = bbox_xywh.clone()
        bbox_tlwh[:, 0] = bbox_xywh[:, 0] - bbox_xywh[:, 2] / 2.0
        bbox_tlwh[:, 1] = bbox_xywh[:, 1] - bbox_xywh[:, 3] / 2.0
        return bbox_tlwh


class DeepSortTorchFeatureExtractor:
    def __init__(self,
                 model_name,
                 device,
                 model_path,
                 ):
        self.extractor = FeatureExtractor(
            model_name=model_name, model_path=model_path, device=str(device)
        )

    def get_features(self, bbox_xywh, ori_img):
        height, width = ori_img.shape[:2]
        im_crops = []
        for box in bbox_xywh:
            x1, y1, x2, y2 = self._xywh_to_xyxy(box, height, width)
            im = ori_img[y1:y2, x1:x2]
            im_crops.append(im)
        if im_crops:
            features = self.extractor(im_crops)
        else:
            features = np.array([])
        return features

    def _xywh_to_xyxy(self, bbox_xywh, height, width):
        x, y, w, h = bbox_xywh
        x1 = max(int(x - w / 2), 0)
        x2 = min(int(x + w / 2), width - 1)
        y1 = max(int(y - h / 2), 0)
        y2 = min(int(y + h / 2), height - 1)
        return x1, y1, x2, y2


class DeepSortTorch:
    def __init__(
            self,
            model_name,
            device,
            model_path,
            max_dist=0.2,
            max_iou_distance=0.7,
            max_age=70,
            n_init=3,
            nn_budget=100,
    ):
        self.feature_extractor = DeepSortTorchFeatureExtractor(
            model_name=model_name,
            model_path=model_path,
            device=str(device))
        self.deep_sort_tracker = DeepSortTorchTracker(
            max_dist=max_dist,
            max_iou_distance=max_iou_distance,
            max_age=max_age,
            n_init=n_init,
            nn_budget=nn_budget)

    def update(self, bbox_xywh, confidences, classes, ori_img, use_yolo_preds=False):
        height, width = ori_img.shape[:2]
        features = self.feature_extractor.get_features(bbox_xywh, ori_img)
        outputs = self.deep_sort_tracker.update(features,
                                                classes,
                                                confidences,
                                                bbox_xywh,
                                                height,
                                                width,
                                                use_yolo_preds)
        return outputs

    def visualize(self, img, tracker_output, confidences, class_names):
        img = self.deep_sort_tracker.visualize_output_image(img, tracker_output, confidences, class_names)
        return img



from deep_utils.utils.logging_utils import log_print


def save_train_val_figures(history, save_path, logger=None):
    import os
    from collections import defaultdict

    import matplotlib.pyplot as plt

    os.makedirs(save_path, exist_ok=True)
    train_val = defaultdict(dict)
    for key, val in history.history.items():
        if key.startswith("val_"):
            train_val[key[4:]]["val"] = val
        else:
            train_val[key]["train"] = val
    for key, values in train_val.items():
        plt.figure(figsize=(10, 8))
        if "train" in values:
            plt.plot(values["train"], label="train")
        if "val" in values:
            plt.plot(values["val"], label="val")
        plt.legend()
        plt.xlabel("epochs")
        plt.ylabel(key)
        plt.title(f"Metric {key}")
        plt.savefig(os.path.join(save_path, key + ".jpg"), dpi=500)
    log_print(logger=logger,
              message="Successfully saved figures to {save_path}!")



from typing import Tuple, Union

import torch


class MemoryUtilsTorch:
    @staticmethod
    def estimate_memory(
            model, sample_input, optimizer_type=torch.optim.Adam, use_amp=False, device: Union[str, int] = 0
    ):
        """Predict the maximum memory usage of the model.
        Args:
            optimizer_type (Type): the class name of the optimizer to instantiate
            model (nn.Module): the neural network model
            sample_input (torch.Tensor): A sample input to the network. It should be
                a single item, not a batch, and it will be replicated batch_size times.
            use_amp (bool): whether to estimate based on using mixed precision
            device (torch.device): the device to use
        """
        # Reset model and optimizer
        model = model.cpu()
        optimizer = optimizer_type(model.parameters(), lr=0.001)
        a = torch.cuda.memory_allocated(device)
        model.to(device)
        b = torch.cuda.memory_allocated(device)
        model_memory = b - a
        # model_input = sample_input.unsqueeze(0).repeat(batch_size, 1)
        model(sample_input.to(device))
        c = torch.cuda.memory_allocated(device)
        if use_amp:
            amp_multiplier = 0.5
        else:
            amp_multiplier = 1
        forward_pass_memory = (c - b) * amp_multiplier
        gradient_memory = model_memory
        if isinstance(optimizer, torch.optim.Adam):
            o = 2
        elif isinstance(optimizer, torch.optim.RMSprop):
            o = 1
        elif isinstance(optimizer, torch.optim.SGD):
            o = 0
        else:
            raise ValueError(
                "Unsupported optimizer. Look up how many moments are"
                + "stored by your optimizer and add a case to the optimizer checker."
            )
        gradient_moment_memory = o * gradient_memory
        total_memory = (
                model_memory
                + forward_pass_memory
                + gradient_memory
                + gradient_moment_memory
        )

        return total_memory

    @staticmethod
    def memory_test(
            model,
            optimizer_type=torch.optim.Adam,
            input_size: Tuple[int, int, int, int] = (1, 3, 224, 224),
            use_amp=False,
            device=0,
            get_extra_info=True,
    ):
        assert device != "cpu", "Only works for cpu"
        sample_input = torch.randn(
            input_size,
            dtype=torch.float32,
        ).to(device)
        max_mem_est = MemoryUtilsTorch.estimate_memory(
            model, sample_input, optimizer_type=optimizer_type, use_amp=use_amp, device=device
        )
        print("Maximum Memory Estimate MB", round(max_mem_est / (1024 ** 2), 2))
        if get_extra_info:
            optimizer = optimizer_type(model.parameters(), lr=0.001)
            print(
                "Beginning mem:",
                torch.cuda.memory_allocated(device),
                "Note - this may be higher than 0, which is due to PyTorch caching. Don't worry too much about this number",
            )
            model.to(device)
            print("After model to device:", torch.cuda.memory_allocated(device))
            for i in range(3):
                print("Iteration", i)
                with torch.cuda.amp.autocast(enabled=use_amp):
                    a = torch.cuda.memory_allocated(device)
                    out = model(
                        sample_input.to(device)
                    ).sum()  # Taking the sum here just to get a scalar output
                    b = torch.cuda.memory_allocated(device)
                print("1 - After forward pass",
                      torch.cuda.memory_allocated(device))
                print("2 - Memory consumed by forward pass", b - a)
                out.backward()
                print("3 - After backward pass",
                      torch.cuda.memory_allocated(device))
                optimizer.step()
                print("4 - After optimizer step",
                      torch.cuda.memory_allocated(device))



import numpy as np
import librosa


class LibrosaUtils:
    @staticmethod
    def stft(audio,
             n_fft,
             n_shift,
             win_length=None,
             window="hann",
             center=True,
             pad_mode="reflect"):
        """
        Computes Short-time Fourier transform for a given audio signal using librosa
        :param audio: audio's shape is [time, n_channel]
        :param n_fft:
        :param n_shift:
        :param win_length:
        :param window:
        :param center:
        :param pad_mode:
        :return:
        """

        if audio.ndim == 1:
            single_channel = True
            audio = np.expand_dims(audio, 1)
        else:
            single_channel = False
        audio = audio.astype(np.float32)

        # audio: [time, channel, freq]
        audio = np.stack([librosa.stft(audio[:, ch],
                                       n_fft=n_fft,
                                       hop_length=n_shift,
                                       win_length=win_length,
                                       window=window,
                                       center=center,
                                       pad_mode=pad_mode, ).T for ch in range(audio.shape[1])], axis=1)
        if single_channel:
            # audio: [time, channel, freq] -> [time, freq]
            audio = audio[:, 0]

        return audio

    @staticmethod
    def stft2log_mel_spectrogram(x_stft,
                                 fs,
                                 n_mels,
                                 n_fft,
                                 fmin=None,
                                 fmax=None,
                                 eps=1e-10):
        """
        Converts STFT to log_mel_spectrogram
        :param x_stft: 
        :param fs: 
        :param n_mels: 
        :param n_fft: 
        :param fmin: 
        :param fmax: 
        :param eps: 
        :return: 
        """
        # x_stft: (time, channel, freq) or (time, freq)
        fmin = 0 if fmin is None else fmin
        fmax = fs / 2 if fmax is None else fmax

        # spc: (time, channel, freq) or (time, freq)
        spc = np.abs(x_stft)
        # mel_basis: (mel_freq, freq)
        mel_basis = librosa.filters.mel(fs, n_fft, n_mels, fmin, fmax)
        # lmspc: (time, channel, mel_freq) or (time, mel_freq)
        lmspc = np.log10(np.maximum(eps, np.dot(spc, mel_basis.T)))

        return lmspc

    @staticmethod
    def log_mel_spectrogram(audio,
                            sample_rate,
                            n_fft,
                            hop_length,
                            win_length,
                            num_mels,
                            mel_fmin,
                            mel_fmax,
                            window="hann",
                            pad_mode="reflect"):
        """
        Computes log mel_spectrogram of an audio signal
        :param audio: 
        :param sample_rate: 
        :param n_fft: 
        :param hop_length: 
        :param win_length: 
        :param num_mels: 
        :param mel_fmin: 
        :param mel_fmax: 
        :param window: 
        :param pad_mode: 
        :return: 
        """
        # Compute STFT
        x_stft = LibrosaUtils.stft(audio,
                                   n_fft=n_fft,
                                   n_shift=hop_length,
                                   win_length=win_length,
                                   window=window,
                                   pad_mode=pad_mode)

        # Compute log-mel-spectrogram
        return LibrosaUtils.stft2log_mel_spectrogram(x_stft,
                                                     fs=sample_rate,
                                                     n_mels=num_mels,
                                                     n_fft=n_fft,
                                                     fmin=mel_fmin,
                                                     fmax=mel_fmax,
                                                     eps=1e-10).T

    @staticmethod
    def trim_silence(auido, margin_db, frame_length=2048, hop_length=256):
        """
        Trims input wav based on margin_db
        :param auido:
        :param margin_db:
        :param frame_length:
        :param hop_length:
        :return:
        """
        return librosa.effects.trim(auido,
                                    top_db=margin_db,
                                    frame_length=frame_length,
                                    hop_length=hop_length)[0]



import math
from typing import Callable


def cosine_reduce(x, total_steps):
    return math.cos((x / total_steps) * (math.pi / 2))


def warmup_cosine(
    warmup_steps, max_lr, total_steps, optimizer_lr=None, initial_lr=1e-6, min_lr=0
) -> Callable:
    """

    How to use it:
    import torch
    scheduler = torch.optim.lr_scheduler.LambdaLR(optimizer, warmup_cosine(...))
    :param warmup_steps:
    :param max_lr:
    :param total_steps:
    :param optimizer_lr: While using huggingface trainer, the output lr is multiplied to the main lr of the optimizer
    which is a bit confusing but if using huggingface and want to make sure to have exact values pass in the optimizer_lr
    value as well so that it will become uneffected.
    :param initial_lr:
    :param min_lr:
    :return:
    >>> epochs = 25
    >>> warmup_steps = epochs//10
    >>> max_lr = 5e-5
    >>> lambda_func = warmup_cosine(warmup_steps, max_lr, epochs)
    >>> for step in range(1, epochs + 1):
    ...     print("lr:", round(lambda_func(step), 6))
    lr: 1.5e-05
    lr: 5e-05
    lr: 5e-05
    lr: 5e-05
    lr: 4.9e-05
    lr: 4.8e-05
    lr: 4.7e-05
    lr: 4.6e-05
    lr: 4.4e-05
    lr: 4.3e-05
    lr: 4.1e-05
    lr: 3.9e-05
    lr: 3.7e-05
    lr: 3.4e-05
    lr: 3.2e-05
    lr: 2.9e-05
    lr: 2.6e-05
    lr: 2.3e-05
    lr: 2e-05
    lr: 1.7e-05
    lr: 1.3e-05
    lr: 1e-05
    lr: 7e-06
    lr: 3e-06
    lr: 0.0
    """
    assert initial_lr < max_lr, "max_lr can't be equal to or less than initial lr"
    assert warmup_steps < total_steps, "warmup_steps can't be larger than total_steps"

    def lambda_func(x):
        if x <= warmup_steps:
            lr = initial_lr + (1 - cosine_reduce(x, warmup_steps)) * (
                max_lr - initial_lr
            )
        else:
            x = x - warmup_steps
            lr = min_lr + cosine_reduce(x, total_steps - warmup_steps) * (
                max_lr - min_lr
            )
            lr = max(min_lr, lr)
        if optimizer_lr is not None and isinstance(optimizer_lr, float):
            return lr / optimizer_lr
        return lr

    return lambda_func



import math
from pathlib import Path
from typing import Union, List

import torch
import torchaudio
from torchaudio import transforms as T

from deep_utils.utils.logging_utils import log_print
from deep_utils.utils.os_utils.os_path import split_extension


class TorchAudioUtils:
    @staticmethod
    def resample(
            wave: Union[str, Path, torch.Tensor],
            sr: int = None,
            resample_rate=16000,
            save=False,
            resampled_path: str = None,
            logger=None,
            verbose=1,
    ) -> Union[str, torch.Tensor]:
        """
        This file is used to resample an audio file or torch tensor
        :param wave:
        :param sr:
        :param resample_rate:
        :param save:
        :param resampled_path:
        :param logger:
        :param verbose:
        :return:
        """

        if isinstance(wave, str) or isinstance(wave, Path):
            waveform, sample_rate = torchaudio.load(wave)
            if save:
                wave_path = (
                    split_extension(wave, suffix=f"_{resample_rate}")
                    if resampled_path is None
                    else resampled_path
                )
        else:
            waveform = wave
            sample_rate = sr
            if save:
                wave_path = resampled_path

        if sr != resample_rate:
            resampler = T.Resample(sample_rate, resample_rate, dtype=waveform.dtype)
            resampled_waveform = resampler(waveform)
        else:
            resampled_waveform = waveform
        if save:
            torchaudio.save(
                wave_path,
                resampled_waveform,
                resample_rate,
                encoding="PCM_S",
                bits_per_sample=16,
            )
            log_print(
                logger,
                f"Successfully resampled and saved wav-file to {wave_path} with {resample_rate} sample rate!",
                verbose=verbose
            )
            return wave_path
        else:
            log_print(
                logger,
                f"Successfully resampled wav-file {waveform.shape} with {resample_rate} sample rate!",
                verbose=verbose
            )
            return resampled_waveform

    @staticmethod
    def split(
            wave, sr, max_seconds: float = 10, min_seconds=1, logger=None, verbose=1
    ) -> List[torch.Tensor]:
        """
        Splits a wave to mini-waves based on input max_seconds. If the last segment's duration is less than min_seconds
        it is combined with its previous segment.
        :param wave:
        :param sr:
        :param max_seconds:
        :param min_seconds:
        :param logger:
        :param verbose:
        :return:
        """
        wave_duration = TorchAudioUtils.get_duration(wave, sr)
        if max_seconds is None or wave_duration < max_seconds:
            return [wave]

        unsqueeze = False
        if len(wave.shape) == 2:
            wave = wave.squeeze(0)
            unsqueeze = True
        n_intervals = math.ceil(wave_duration / max_seconds)
        waves = []

        for interval in range(n_intervals):
            s = (interval * max_seconds) * sr
            e = ((interval + 1) * max_seconds) * sr
            w = wave[s:e]
            w_duration = TorchAudioUtils.get_duration(w, sr)
            w = w.unsqueeze(0) if unsqueeze else w
            if w_duration < min_seconds:
                waves[-1] = torch.concat([waves[-1], w], dim=1)
            else:
                waves.append(w)
        if len(waves) > 1:
            log_print(
                logger,
                f"Successfully split input wave to {len(waves)} waves!",
                verbose=verbose,
            )
        return waves

    @staticmethod
    def get_duration(wave, sr):
        """
        Computing the duration of a wav array
        :param wave:
        :param sr:
        :return: duration in seconds
        """
        if len(wave.shape) == 2:
            wave = wave.squeeze(0)
        return wave.shape[0] / sr

    @staticmethod
    def get_file_duration(wav_path: Union[Path, str]):
        """
        Computing the duration of a wav file
        :param wav_path:
        :return: duration in seconds
        """
        speech_array, sampling_rate = torchaudio.load(wav_path)
        duration_in_seconds = speech_array.shape[1] / sampling_rate
        return duration_in_seconds



import os

from deep_utils.utils.logging_utils import log_print
from deep_utils.utils.os_utils.os_path import split_extension


class ModelCheckPointTorch:
    def __init__(
        self,
        model_path,
        model,
        monitor="min",
        save_best_only=True,
        overwrite=True,
        save_last=True,
        loss=None,
        optimizer=None,
        scheduler=None,
        static_dict=None,
        monitor_val=None,
        logger=None,
        verbose=True,
    ):
        """

        :param model_path:
        :param model:
        :param monitor:
        :param save_best_only:
        :param overwrite:
        :param verbose:
        :param save_last:
        :param loss:
        :param optimizer:
        :param scheduler:
        :param static_dict:
        :param monitor_val: monitoring value. For starting the checkpointer from a saved model,
         by passing the monitoring value you preserve the best model.
        """
        self.overwrite = overwrite
        self.model_path = model_path
        self.monitor_val = float("-inf") if monitor == "max" else float("inf")
        self.monitor_val = self.monitor_val if monitor_val is None else monitor_val
        self.save_best_only = save_best_only
        self.model = model
        self.loss = loss
        self.optimizer = optimizer
        self.scheduler = scheduler
        self.static_dict = static_dict
        self.monitor = monitor
        self.epoch = 0
        self.verbose = verbose
        self.save_last = save_last
        self.logger = logger
        os.makedirs(os.path.split(model_path)[0], exist_ok=True)
        log_print(
            self.logger,
            f"Initializing checkpointer with monitor-mode:{monitor} from monitor-val:{monitor_val}",
            verbose=verbose,
        )

    def __call__(self, monitor_val, **kwargs):
        self.epoch += 1
        if self.save_best_only:
            trigger = False
            if self.monitor == "min" and monitor_val < self.monitor_val:
                self.monitor_val = monitor_val
                trigger = True
            elif self.monitor == "max" and monitor_val > self.monitor_val:
                self.monitor_val = monitor_val
                trigger = True

            if self.save_last:
                last_path = split_extension(self.model_path, suffix="_last")
                self._save(last_path, print_=False,
                           monitor_val=monitor_val, **kwargs)

            if self.overwrite:
                best_path = split_extension(self.model_path, suffix="_best")
            else:
                best_path = split_extension(
                    self.model_path, suffix="_" + str(self.epoch)
                )

            if trigger:
                self._save(
                    best_path, print_=self.verbose, monitor_val=monitor_val, **kwargs
                )
        else:
            model_path = split_extension(
                self.model_path, suffix="_" + str(self.epoch))
            self._save(
                model_path, print_=self.verbose, monitor_val=monitor_val, **kwargs
            )

    def _save(self, model_path, print_, **extra):
        import torch

        save_dict = self.static_dict if self.static_dict is not None else dict()
        save_dict.update(extra if extra is not None else dict())
        save_dict["model_state_dict"] = self.model.state_dict()
        self._add_file(save_dict, "optimizer", self.optimizer.state_dict())
        self._add_file(save_dict, "scheduler", self.scheduler.state_dict())
        self._add_file(save_dict, "loss", self.loss)
        torch.save(save_dict, model_path)
        if print_:
            if "monitor_val" in extra:
                log_print(
                    self.logger,
                    f'monitor-val: {extra["monitor_val"]}, model is saved in {model_path}',
                    verbose=self.verbose,
                )
            else:
                log_print(
                    self.logger, f"model is saved in {model_path}", verbose=self.verbose
                )

    @staticmethod
    def _add_file(dict_, name, file):
        if file is not None:
            dict_[name] = file
        return dict_



from deep_utils.utils.box_utils.boxes import Box, Point
from deep_utils.utils.lib_utils.download_utils import download_decorator
from deep_utils.utils.lib_utils.lib_decorators import (
    expand_input,
    get_elapsed_time,
    get_from_config,
    rgb2bgr,
)
from deep_utils.vision.face_detection.main import FaceDetector

from .config import Config


class HaarcascadeCV2FaceDetector(FaceDetector):
    def __init__(self, **kwargs):
        super().__init__(
            name=self.__class__.__name__,
            file_path=__file__,
            download_variables=(
                "haarcascade_frontalface",
                "haarcascade_eye",
                "haarcascade_nose",
                "landmarks",
            ),
            **kwargs
        )
        self.config: Config

    @download_decorator
    def load_model(self):
        import cv2

        face_detector = cv2.CascadeClassifier(
            self.config.haarcascade_frontalface)
        nose_detector = cv2.CascadeClassifier(self.config.haarcascade_nose)
        eye_detector = cv2.CascadeClassifier(self.config.haarcascade_eye)
        landmark_detector = cv2.face.createFacemarkLBF()
        landmark_detector.loadModel(self.config.landmarks)
        self.model = dict(
            face=face_detector,
            nose=nose_detector,
            eye=eye_detector,
            landmarks=landmark_detector,
        )

    @get_elapsed_time
    @expand_input(3)
    @get_from_config
    @rgb2bgr("gray")
    def detect_faces(
        self,
        img,
        is_rgb,
        scaleFactor=None,
        minNeighbors=None,
        minSize=None,
        maxSize=None,
        flags=None,
        get_landmarks=True,
        get_nose=True,
        get_eye=True,
        get_time=False,
    ):
        boxes, landmarks_, confidences, eye_poses, nose_poses = [], [], [], [], []
        for image in img:
            faces = self.model["face"].detectMultiScale(
                image,
                scaleFactor=scaleFactor,
                minNeighbors=minNeighbors,
                minSize=minSize,
                maxSize=maxSize,
                flags=flags,
            )
            boxes.append(
                Box.box2box(
                    faces,
                    in_source="CV",
                    to_source="Numpy",
                    in_format="XYWH",
                    to_format="XYXY",
                )
            )
            if get_nose:
                nose_pos = self.model["nose"].detectMultiScale(
                    image,
                    scaleFactor=scaleFactor,
                    minNeighbors=minNeighbors,
                    minSize=minSize,
                    maxSize=maxSize,
                    flags=flags,
                )
                nose_pos = Point.point2point(
                    nose_pos, in_source="CV", to_source="Numpy"
                )
                nose_poses.append(nose_pos)
            if get_eye:
                eye_pos = self.model["eye"].detectMultiScale(
                    image,
                    scaleFactor=scaleFactor,
                    minNeighbors=minNeighbors,
                    minSize=minSize,
                    maxSize=maxSize,
                    flags=flags,
                )
                eye_pos = Point.point2point(
                    eye_pos, in_source="CV", to_source="Numpy")
                eye_poses.append(eye_pos)
            if len(faces) != 0 and get_landmarks:
                _, landmarks = self.model["landmarks"].fit(image, faces)
                landmarks = [
                    Point.point2point(
                        face_landmarks[0].tolist(), in_source="CV", to_source="Numpy"
                    )
                    for face_landmarks in landmarks
                ]
                landmarks_.append(landmarks)
        return dict(
            boxes=boxes,
            confidences=confidences,
            landmarks=landmarks_,
            eye_poses=eye_poses,
            nose_poses=nose_poses,
        )



from pyannote.audio import Pipeline
from deep_utils.utils.logging_utils.logging_utils import log_print, get_logger
from deep_utils.audio.audio_utils.torchaudio_utils import TorchAudioUtils
import torchaudio


class PyannoteAudioDiarization:
    def __init__(self, device="cpu", min_duration_on=1, max_duration_on=10, logger=None,
                 checkpoint_path="pyannote/speaker-segmentation", use_auth_token=None):
        self.max_duration_on = max_duration_on
        # self.diarize_model = SpeakerSegmentation("pyannote/segmentation")
        self.diarize_model = Pipeline.from_pretrained(checkpoint_path=checkpoint_path, use_auth_token=use_auth_token)
        self.diarize_model._segmentation.model.to(device)
        self.diarize_model._segmentation.device = device
        self.diarize_model.instantiate({
            # onset/offset activation thresholds
            "onset": 0.5, "offset": 0.5,
            # remove speech regions shorter than that many seconds.
            "min_duration_on": min_duration_on,
            # fill non-speech regions shorter than that many seconds.
            "min_duration_off": 0
        })

        log_print(logger, "Successfully Created Speaker Diarization!")

    def infer(self, wave, sr, return_segments=True, logger=None, verbose=1):
        model_input = dict({"waveform": wave, "sample_rate": sr})
        wave_note = f"wave-shape: {wave.shape}, sr: {sr}"
        out = self.diarize_model(model_input)
        indices, audio_segments = [], []
        if return_segments:
            for speech_region, _, speaker in out.itertracks(yield_label=True):
                w = wave[0][int(speech_region.start * sr):int(speech_region.end * sr)].unsqueeze(0)
                ws = TorchAudioUtils.split(w, sr=sr, max_seconds=self.max_duration_on, logger=logger, verbose=verbose)
                indices.extend([int(speaker.split('_')[-1]) for _ in range(len(ws))])
                audio_segments.extend(ws)
            log_print(logger, f"Successfully diarized wav-file: {wave_note} to {len(audio_segments)} segments")
            return indices, audio_segments
        else:
            output = [(speaker, (speech_region.start, speech_region.end)) for speech_region, _, speaker in
                      out.itertracks(yield_label=True)]
            log_print(logger, f"Successfully diarized wav-file: {wave_note} to {len(audio_segments)} segments")
            return output

    def infer_file(self, wave_path, logger=None):
        wave, sr = torchaudio.load(wave_path)
        indices, audio_segments = self.infer(wave=wave, sr=sr, logger=logger)
        return indices, audio_segments

    def infer_group(self, audio_segments, sr, logger=None, verbose=1):
        res_segments, suffix = [], []
        for i, audio_segment in enumerate(audio_segments):
            indices, segments = self.infer(audio_segment, sr=sr, logger=logger)

            if indices:
                for j, (index, segment) in enumerate(zip(indices, segments)):
                    res_segments.append(segment)
                    suffix.append(f"{i:03}_{j:02}_S{index:02}")
            else:
                res_segments.append(audio_segment)
                suffix.append(f"{i:03}_{0:02}_S{0:02}")
        log_print(logger, f"Successfully diarized {len(suffix)} samples", verbose=verbose)
        return res_segments, suffix


if __name__ == '__main__':
    a = PyannoteAudioDiarization()
    a.infer(logger=1)



try:
    from deep_utils.dummy_objects.vision.object_tracker import DeepSortTorch
    from deep_utils.vision.object_tracker.deep_sort import DeepSortTorch
except:
    pass


try:
    from deep_utils.dummy_objects.vision.object_tracker import DeepSortTorchFeatureExtractor
    from deep_utils.vision.object_tracker.deep_sort import DeepSortTorchFeatureExtractor
except:
    pass

try:
    from deep_utils.dummy_objects.vision.object_tracker import DeepSortTorchTracker
    from deep_utils.vision.object_tracker.deep_sort import DeepSortTorchTracker
except:
    pass


def get_y_pred_true(model, test_gen, return_numpy=True):
    """
    Computes y_pred and returns its corresponding t_true as well
    Args:
        model:
        test_gen:
        return_numpy:

    Returns: y_pred and y_true

    """
    import numpy as np

    print("[INFO] Making  prediction ")
    y_pred, y_true = [], []
    for x, y in test_gen:
        predictions = model.predict(x, batch_size=1)
        temp = np.argmax(predictions, axis=1)
        y = np.argmax(y, axis=1)
        y_pred.extend(temp)
        y_true.extend(y)
    if return_numpy:
        y_true, y_pred = np.array(y_true), np.array(y_pred)
    return y_pred, y_true



from .cutmix_tf import CutMixTF

try:
    from deep_utils.augmentation.cutmix.cutmix_torch import CutMixTorch
    from deep_utils.dummy_objects.augmentation.cutmix import CutMixTorch
except:
    pass



import inspect
import logging
import os
import sys
from functools import wraps
from pathlib import Path
from typing import Callable, Union

LOGGING_FORMATS = {
    "basic": "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    "func_name": "[%(filename)s:%(lineno)s - %(funcName)20s() ] %(message)s",
}


def start_end_logger_decorator(func) -> Callable:
    """
    Logs the start and end of a func
    :param func: The func that will be measured
    :return:
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        logger_ = kwargs.get("logger", None)
        verbose = kwargs.get("verbose", 1)
        file_path = inspect.getfile(func)
        func_name = func.__name__
        log_print(
            logger_,
            f"{file_path} -- func: {func_name} --> message: Starting...",
            verbose=verbose,
            get_func_name=False,
        )
        results = func(*args, **kwargs)
        log_print(
            logger_,
            f"{file_path} -- func: {func_name} --> message: Done...",
            verbose=verbose,
            get_func_name=False,
        )
        return results

    return wrapper


def get_logger(
    name: str,
    log_path: Union[str, Path, None] = None,
    remove_previous_handlers=True,
    logging_format: str = "basic",
) -> logging.Logger:
    """
    Creates a logger for a given name,
    :param name: The name that logger will be created for
    :param log_path: Where logger information will be saved
    :param remove_previous_handlers: If set to true, removes previous handlers to prevent repeated prints
    :param logging_format: What format to use, default is basic. Check LOGGING_FORMATS dict for available formats.
    :return:
    """
    logger = logging.getLogger(name)
    formatter = logging.Formatter(LOGGING_FORMATS[logging_format])

    if log_path:
        log_dir, file_name = os.path.split(log_path)
        if log_dir:
            os.makedirs(log_dir, exist_ok=True)
        file_handler = logging.FileHandler(log_path)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)
    else:
        stream_handler = logging.StreamHandler(sys.stdout)
        stream_handler.setFormatter(formatter)
        logger.addHandler(stream_handler)

    if remove_previous_handlers:
        logger.handlers = logger.handlers[-1:]
    logger.setLevel(logging.INFO)
    print(
        f"[INFO] Successfully created logger for {name}"
        + (f" in {log_path}" if log_path else "")
    )
    return logger


def func_log(message, roll_back=1):
    """Automatically log the current function details."""

    # Get the previous frame in the stack, otherwise it would
    # be this function!!!
    code_f_back = inspect.currentframe()
    for _ in range(roll_back):
        code_f_back = code_f_back.f_back
    func = code_f_back.f_code
    # Dump the message + the name of this function to the log.
    message = f"message: {message} --> file: {func.co_filename} -- line: {func.co_firstlineno} -- func: {func.co_name}"
    return message


def log_print(
    logger: Union[None, logging.Logger],
    message: str,
    log_type="info",
    verbose=1,
    get_func_name=True,
    roll_back=2,
):
    """
    Logs the input messages with the given log_type. In case the logger object is not provided, prints the message.
    :param logger:
    :param message:
    :param log_type:
    :param verbose: whether to print/log
    :param get_func_name: whether to include func-name in the message
    :param roll_back: roll back used for func-log. This number indicates how many functions should `func-log` trace back
    to get to the main func
    :return:
    """
    if get_func_name:
        message = func_log(message, roll_back=roll_back)
    if log_type == "info":
        if logger is not None and isinstance(logger, logging.Logger):
            logger.info(message)
        else:
            if verbose:
                print(f"[INFO] {message}")
    elif log_type == "error":
        if logger is not None and isinstance(logger, logging.Logger):
            logger.error(message)
        else:
            if verbose:
                print(f"[ERROR] {message}")
    else:
        if verbose:
            print(f"[ERROR] log_type: {log_type} is not supported")
        raise ValueError("[ERROR] log_type: {log_type} is not supported")


def value_error_log(logger: Union[None, logging.Logger], message: str):
    """
    generates a value error and logs it!
    :param logger:
    :param message:
    :return:
    """
    log_print(logger, message, log_type="error")
    raise ValueError(message)


def save_params(param_path, args, logger=None):
    """
    Save the arguments in the given path.
    Args:
        param_path:
        args:
        logger: If provided the message will be logged unless will be printed to console!

    Returns:

    """
    log_print(logger, f"Saving params!")
    with open(param_path, mode="w") as f:
        arguments = vars(args)
        for key, val in arguments.items():
            f.write(f"{key} {val}\n")
    log_print(logger, f"Params are successfully saved in {param_path}!")


def get_conf_matrix(
    class_name_map,
    y_pred,
    y_true,
    save_path=None,
    conf_csv_name="conf_matrix.csv",
    conf_jpg_name="conf_matrix.jpg",
    logger=None,
):
    """
    Computes config matrix and saves the csv and jpg file if the save_path is provided!
    Args:
        class_name_map:
        y_pred:
        y_true:
        save_path:
        conf_csv_name:
        conf_jpg_name:
        logger: If provided the message will be logged unless will be printed to console!

    Returns: configuration matrix

    """
    import matplotlib.pyplot as plt
    import pandas as pd
    import seaborn as sns
    from sklearn.metrics import confusion_matrix

    conf_matrix = confusion_matrix(y_true, y_pred)
    df_cm = pd.DataFrame(
        conf_matrix,
        index=list(class_name_map.keys()),
        columns=list(class_name_map.keys()),
    )
    plt.figure(figsize=(10, 7))
    sns.heatmap(df_cm, annot=True, fmt="g")
    plt.xlabel("")
    plt.ylabel("")
    if save_path is not None:
        df_cm.to_csv(os.path.join(save_path, conf_csv_name))
        plt.savefig(os.path.join(save_path, conf_jpg_name))
    log_print(logger, "Confusion matrix is successfully generated!")
    return conf_matrix


def get_cls_report(y_pred, y_true, save_path=None, logger=None):
    """
    Generate classification report and saves them if the save_path is provided!
    Args:
        y_pred:
        y_true:
        save_path:
        logger: If provided the message will be logged unless will be printed to console!

    Returns:

    """
    from sklearn.metrics import classification_report

    report = classification_report(y_true, y_pred)
    if save_path:
        with open(save_path, mode="w") as f:
            f.write(report)
    log_print(logger, "Successfully generated classification report")
    return report


if __name__ == "__main__":
    l = get_logger("p", log_path="log.log")
    log_print(l, "wow", verbose=1)
